/*
** Module   : GenericFrame.java
** Abstract : implementation of UI frames using dynamic proxy feature
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20051026   @23332 Created initial version.
** 002 NVS 20060114   @24058 Extended hide() methods.
** 003 GES 20060115   @23919 Fixes to getters/setters and view.
** 004 NVS 20060116   @24059 hide() and view() methods change visibility
**                           flag in frame config instance.
** 005 GES 20060116   @23946 Implemented conversion of 1-based to 0-based
**                           positions.
** 006 NVS 20060125   @24060 up() methods were calling down() in LT.
** 007 NVS 20060125   @24069 Removed all traces of DownFrame class.
** 008 SIY 20060125   @24117 Frame manager support code.
** 009 GES 20060203   @24260 Fixed getter for isXYZ() methods. Added
**                           widget init code to cache the ID. Rewrote
**                           ID and widget lookup methods to be simpler 
**                           and faster.  Added some new interfaces for
**                           ID and widget access.  Simplified the data
**                           storage approach. Rewrote call sites to
**                           use common, faster implementations.
** 010 GES 20060204   @24270 Added display/update/set/prompt-for methods.
** 011 GES 20060208   @24383 Modifications for redirected I/O and using
**                           the UI statements on streams instead of the
**                           terminal.
** 012 SIY 20060210   @24512 Added getSelection() method.                   
** 013 SIY 20060213   @24582 Fix for the DialogBox support.
** 014 SIY 20060216   @24609 Removed unnecessary nested scopes handling.
** 015 SIY 20060222   @24630 Minor change in CHOOSE handling.
** 016 NVS 20060223   @24665 Within the UPDATE processing, the internal
**                           block should not catch the END condition.
** 017 GES 20060223   @24684 Final redirected output support.
** 018 NVS 20060224   @24727 High level I/O operations pass their codes
**                           into the WaitList so that the client has this
**                           information. Used for END-ERROR handling.
** 019 NVS 20060224   @24732 Since update() creates its own inner block,
**                           infinite loop protection should be disabled
**                           twice - on entry into update() and in the
**                           inner block, which is done in LT.waitFor().
** 020 GES 20060226   @24759 Added support for report headers/footers by
**                           honoring page-top and page-bottom attributes.
** 021 SIY 20060226   @24766 Widget arrays handling throws an exception
**                           if appropriate variable is undefined.
**                           Minor cleanups.
** 022 GES 20060228   @24777 Support prompt-for editing blocks.
** 023 GES 20060228   @24784 Support for a special behavior of the
**                           screen-value attribute and the input
**                           function.
** 024 GES 20060301   @24800 New feature to expose proper ASSIGN semantic
**                           in the case where the screen-buffer has no
**                           value for a given widget.
** 025 GES 20060303   @24877 Fixed error processing and missing error #.
** 026 NVS 20060306   @24905 WaitList does not need default events, they
**                           are recognized by ThinClient anyway.
** 027 NVS 20060308   @24940 ScreenBuffer is instantiated with state flags
** 028 NVS 20060309   @24963 Using a differential ScreenBuffer.
** 029 NVS 20060314   @25073 Resetting all widgets to the uninitialized
**                           state with every frame iteration.
** 030 SIY 20060316   @25132 Added method to retrieve widget name for the
**                           specified widget.
** 031 GES 20060317   @25084 Header support (includes some code written
**                           by SIY).
** 032 SIY 20060317   @23133 Merged support for delayed browse init.
** 033 SIY 20060318   @25134 Fixes for the header support.
** 034 NVS 20060321   @25177 Added getValidationSet() method.
** 035 GES 20060322   @25192 Added new signatures for title processing.
** 036 SIY 20060323   @25221 Fixed passing of widget values when list of
**                           widgets is provided explicitly.
** 037 GES 20060401   @25300 Added display sigs for the case where there
**                           are no data containing widgets in a frame.
** 038 SIY 20060401   @25306 Temporary workaround for the initial widget
**                           display.
** 039 GES 20060404   @25346 Moved enable/disable to a frame element
**                           approach.
** 040 SIY 20060404   @25356 Support for the text group.
** 041 SIY 20060409   @25434 Fixed handling of null items in the frame
**                           elements array.
** 042 GES 20060413   @25514 Added setScreenValue() method.
** 043 SIY 20060414   @25537 Added setVisible() methods.
** 044 SIY 20060415   @25549 Added enableExcept()/disableExcept() for 
**                           single widget.
** 045 GES 20060417   @25557 Rewrote setScreenValue() to handle all BDT
**                           types properly.
** 046 ECF 20060424   @25674 Added setScrollable() variant which accepts a
**                           logical.
** 047 SIY 20060424   @25685 Added getter/setter for parent and asWidget()
**                           method.
** 048 GES 20060426   @25731 Added underline() methods.
** 049 GES 20060427   @25820 Added setStatisticsColumn() method.
** 050 GES 20060428   @25868 Added accumulator support to display().
** 051 GES 20060502   @25931 Added possibleValidFrameId().
** 052 NVS 20060503   @25946 Repeating calls to setTitle have no effect.
**                           Title is set with the first call supplying a
**                           non-empty title string.
** 053 ECF 20060503   @25956 Completed first pass accumulator support for
**                           display aggregators.
** 054 NVS 20060504   @25978 setTitle() calls from the business logic are
**                           now called setDynamicTitle(). The original
**                           set is preserved for use in the frame
**                           creation logic.
** 055 ECF 20060504   @25980 Fix in AggregateHandler. List of full result
**                           aggregator elements was not being sorted.
** 056 GES 20060505   @25988 Detect when an error condition was caused by
**                           the END-ERROR key and rethrow the exception
**                           rather than retry in that case.
** 057 NVS 20060509   @26037 Implemented proper line spacing for totals
**                           in aggregated output.
** 058 SIY 20060508   @26038 Added notification of the client about
**                           opening of the frame scope.
** 059 GES 20060511   @26068 Reset the changed flags in setScreenBuffer()
**                           so that future getFrameBuffer() calls don't
**                           resend data as changed that has already been
**                           sent.  Widget state is modified on the 
**                           client when setting "changed" values and this 
**                           causes deviations from how Progress works.                               
** 060 GES 20060511   @26069 Fix for continueEditing() which was
**                           improperly disabling/copying back to the
**                           screen buffer on the first pass through the
**                           editing loop which doesn't have a pending GO. 
**                           Thus the logic was inverted when it should
**                           not have been.
** 061 NVS 20060511   @26089 Reimplemented conditional down() and up()
**                           calls which were simply short-circuited to
**                           down(1) and up(1) calls. Changed the 
**                           aggregated display accordingly. 
**                           Fixed scroll().
** 062 GES 20060517   @26195 Match new Finalizable interface requirement.
** 063 GES 20060523   @26422 Reworked internals to ensure that an abnormal
**                           exit from an editing block will still handle
**                           cleanup.
** 064 GES 20060523   @26450 SET/UPDATE/PROMPT-FOR must disable even when
**                           there is an abnormal exit.
** 065 GES 20060531   @26826 SET/UPDATE/PROMPT-FOR and editing blocks
**                           must force the focus to change to the first
**                           listed widget.
** 066 NVS 20060601   @26894 getFrameLine() now accounts for any pending
**                           conditional UP/DOWNs as it should.
** 067 SIY 20060601   @26915 promptFor() should clear messages at the end.
** 068 GES 20060612   @27120 Removed id list form of getFrameBuffer() as
**                           it should not be used because no matter what
**                           widgets are specified in a display, view,
**                           prompt-for, set or update, any changed data
**                           in the screen buffer is made visible on the
**                           client.
** 069 SIY 20060615   @27304 Clear frame buffer in clearAll().
** 070 GES 20060616   @27393 Clear frame buffer in clear().
** 071 SIY 20060622   @27536 Added some code for frame destroying mechanism
**                           (not functional yet).
**                           Frame name is stored in the config (can be
**                           overridden by frame initialization code).
** 072 GES 20060627   @27601 Added new setDown() signature to provide
**                           fully dynamic down frame control. Updated
**                           screen buffer processing to honor this new
**                           value.
** 073 NVS 20060627   @27608 Fixed all unconditional forms of up() and
**                           down(). Now they incorporate the function of
**                           conditionalUpDown() method but without making
**                           an extra trip to the client. Not only does it
**                           save some time, it fixes some inconsistency
**                           and eliminates unneeded pausing.
** 074 SIY 20060629   @27696 Fixed up() handling. Some refactoring for 
**                           better code transparency and maintainability.
** 075 GES 20060705   @27806 Moved setColors() from LogicalTerminal and
**                           added batching and conditionalUpDown.
** 076 SIY 20060705   @27838 Removed old workaround in scroll() which is
**                           not necessary anymore and causes regressions.
** 077 GES 20060707   @27906 Overhaul of setDynamicTitle to support a
**                           full callback model. The callback is stored
**                           and the resulting text is included in the
**                           screen buffer every time it is sent down to
**                           the thin client.
** 078 SIY 20060717   @28083 Call batch(close) after setUp().
** 079 GES 20060728   @28218 Moved conditional up/down processing to
**                           a state synch approach to avoid an extra
**                           round trip to the client for each view()
**                           or other language statement that must
**                           honor such things.
** 080 SIY 20060726   @28234 Clear the frame state when it goes of of 
**                           scope. Refactored frame hiding algorithm. 
** 081 GES 20060731   @28237 Minor changes to enable deferring of the
**                           destroyFrame() call.
** 082 SIY 20060731   @28250 Frame clearing now maintained more carefully.
** 083 SIY 20060801   @28334 Clear pending up/down on retry.
** 084 GES 20060803   @28378 Removed unnecessary cast to ComponentEntity.
** 085 GES 20060804   @28429 Shifted the aggregation handler's iterate
**                           processing to allow for the fact that iterate
**                           is only called at the top of the block. The
**                           previous version depended on being called at
**                           the bottom of the loop.
** 086 SIY 20060807   @28451 Changes in resetting of the frame.
** 087 GES 20060809   @28498 Added proper temporary terminal redirection
**                           to the up/down methods which take an explicit
**                           stream parameter.
** 088 SIY 20060810   @28517 Destroy frame when leaving current procedure
**                           or function, not only execute().                      
** 089 SIY 20060810   @28536 Partially rolled back #28517 to fix 
**                           regressions.
** 090 GES 20060816   @28626 Fix class cast problem by properly 
**                           dereferencing a stream's remote ID (one of
**                           the 2 cases was not being handled properly).
** 091 SIY 20060823   @28761 Added support for displaying static widgets
**                           in SET/UPDATE/PROMPT-FOR.
** 092 SIY 20060825   @28894 Protect from frame buffer resetting if frame 
**                           is not displayed yet.
** 093 NVS 20060830   @29021 Consolidated enable() and view() functions in
**                           the enable() method. No separate view() call
**                           is needed after call to enable().
** 094 GES 20060908   @29373 Avoid assignment of static widget values in
**                           assignOnly and widgetOnly cases.
** 095 GES 20060908   @29622 null as a stream reference must be treated
**                           as the terminal (remote stream id == -1).
** 096 GES 20060915   @29623 Honor overridden format strings for a given
**                           value to be placed in the screen buffer.
** 097 GES 20060918   @29663 Database fields now have J2SE-compatible
**                           variable names which is incompatible with the
**                           simplistic approach of "always decapitalize" 
**                           (the first char of a variable name). Now we 
**                           try the Java beans approach if the simplistic
**                           approach fails.
** 098 GES 20060919   @29687 Conversion is needed for cases where @ base
**                           field data is read from the screen buffer
**                           into a BDT of a different type.
** 099 GES 20060919   @29692 Data type conversion is needed in the
**                           assignScreenValue() method.  Now we use the
**                           common getter() code to handle this.
** 100 GES 20060920   @29721 Properly handle unknown literals as data
**                           type overrides in getter().
** 101 GES 20060926   @30025 Fixed TEXT group handling.
** 102 NVS 20061005   @30166 Added a full implementation of FRAME-VALUE
**                           statement as setFrameValue() method.
** 103 NVS 20061006   @30202 Added backing code for retrieving original
**                           widget names required by FRAME-FIELD.
** 104 GES 20061010   @30261 Implemented full support for shared frames.
** 105 GES 20061031   @30825 Safety code to ensure that in any failure
**                           a temporary terminal redirection is removed
**                           before continuing.  This gives us a cleaner
**                           abend and may fix some issues in the case
**                           where a failure is really a Progress 
**                           condition like ENDKEY or ERROR.
** 106 GES 20061031   @30830 Safety code to translate unknown to the
**                           proper type before updating the screen
**                           buffer.
** 107 GES 20061101   @30891 Deregister headers at end of block to which
**                           the frame is scoped.
** 108 SIY 20061102   @30899 Do not add Skip if there are no widgets in
**                           the frame body.
** 109 GES 20061103   @30930 Don't override visibility flag of a shared
**                           frame when we import that frame.
** 110 SIY 20061108   @31038 Support for special case of accessing array 
**                           widgets.
** 111 GES 20061120   @31294 Updated for support of dynamic row/column 
**                           and down.
** 112 GES 20061127   @31409 Added a shared flag to destroyFrame() to
**                           ensure that more deeply nested frames don't
**                           exit scope too soon.
** 113 SIY 20061128   @31459 Added scope nesting counter and excluded 
**                           nested scopes of shared frames from regular
**                           scope processing. 
** 114 NVS 20061129   @31473 Frame entering top level scope signals the
**                           fact to the client. Scope nesting level was
**                           decremented in a wrong place, fixed.
** 115 GES 20061206   @31595 Use new promptFor() helper.
** 116 GES 20061212   @31664 Eliminated most switchOutput() bracketing by 
**                           using new client exports that implement the
**                           temporary redirection in-line on the client
**                           rather than driving it via separate calls
**                           from the server.  This should have a major
**                           performance improvement for some forms of
**                           redirected output generation.
** 117 SIY 20061225   @31747 Renamed setUp() to setup(). 
** 118 SIY 20070116   @31843 Changed location of registration in TM for 
**                           top level finalization processing.
** 119 SIY 20070117   @31848 New frame ID assignment scheme.
** 120 NVS 20070118   @31887 Using WidgetId class and new trigger registry
**                           management based on EventList class. WaitList
**                           class is no longer in use.
** 121 SIY 20070118   @31899 More changes in finalization and shared
**                           frame handling.
** 122 SIY 20070124   @31972 Removed redundant registration of frame in TM.
** 123 SIY 20070306   @32282 Do not deregister frame in LogicalTerminal. 
** 124 SIY 20070314   @32399 Finish setup of browse widgets even in 
**                           widget-specific enable() call.
** 125 SIY 20070329   @32658 Fixed order of IDs returned by remainingIds().
** 126 SIY 20070409   @32874 Do not call down() in iterate() for regular 
**                           frames.
** 127 NVS 20070502   @33358 Made one of the getter() methods package
**                           private from private as it's used from
**                           LogicalTerminal.
** 128 NVS 20070507   @33418 Establishing proper connection between this
**                           instance of GenericFrame and an instance of
**                           FrameWidget.
** 129 SIY 20070521   @33751 Call conditionalUpDown() before calling 
**                           LogicalTerminal.refreshFrameWidget().
**                           Added signalling of closing scope.
** 130 SIY 20070601   @33925 Added display(BrowseWidget).
** 131 SIY 20070711   @34428 Fixed shared frame support.
** 132 NVS 20070716   @34537 Rerouted methods using unnamed streams to
**                           call matching methods using named streams.
**                           Added idsFromWidgets() method.
** 133 NVS 20070724   @34676 NOT THE FINAL CHANGE. All view(), display()
**                           calls pass the flag that tells the nature of
**                           the call to the workers. 
** 134 NVS 20070801   @34750 An attempt to fix conditional UP and DOWN in
**                           redirected mode.
** 135 NVS 20070807   @34797 DOWN for a non-down frame in redirected mode
**                           is a meaningful operation. It flushes the
**                           terminal buffer and inserts the required
**                           number of new lines into the stream.
** 136 NVS 20070808   @34818 Fixed display of page-top frames, which has
**                           been regressed recently.
** 137 NVS 20070808   @34822 display() call without parameters translates
**                           display(empty array) even in redirected mode.
** 138 NVS 20070817   @34879 Fixed display of page element frames without
**                           frame elements.
** 139 NVS 20070829   @34979 Added private method isRedirected(). Reworked
**                           implementations of up() and down() families
**                           of methods.
** 140 NVS 20070907   @35005 Fixed a bug in downWorker which affected up()
**                           method.
** 141 SIY 20070913   @35064 Fixed handling of LASTKEY when reading data
**                           from stream (redirected command/file).
**                           Ref: 35057-35062,35065
** 142 SIY 20070924   @35196 Support for new line feeds handling approach.
**                           Ref: 35198
** 143 NVS 20070925   @35208 Conditional down in redirected mode is not
**                           delayed only if the target frame is realized.
** 144 BRJ 20071003   @35360 Protected stopEditingMode from being called
**                           when startEditingMode was aborted due to no
**                           widgets being present.
** 145 NVS 20071102   @35764 Conditional down in interactive mode issued
**                           right after a redirected display for the same
**                           frame acts like a redirected down.
** 146 SIY 20071121   @35931 Support for Frame 'need pause' flag 
**                           maintenance.
**                           Ref: #35928
** 147 CA  20071122   @35949 Added setRetain(int) implementation. 
**                           Added RETAIN n and SCROLL n behavior for the
**                           down/up statements in method 
**                           downWorker(int,boolean). 
**                           Modified up(int) method to call method
**                           downWorker(int,boolean).      
** 148 CA  20071126   @36015 Replaced all (redirOut != null) conditions 
**                           with (redirOut!=null && redirOut.isOut()) 
**                           for the field GenericFrame.redirOut.      
** 149 SIY 20071126   @36038 Fix for up() handling in DOWN frames without 
**                           RETAIN and SCROLL options. 
**                           Some refactoring and cleanup.
**                           Ref: #36034
** 150 CA  20071129   @36077 Added the Progress caching behavior (for down 
**                           frames) for conditional up/down statements.
**                           Ref: #36076
** 151 NVS 20071203   @36138 Uncommented a condition in a down() call in
**                           iterate() which has been commented out by
**                           mistake
** 152 NVS 20071206   @36232 Aggregation handler is no longer scoped to
**                           the nearest loop. but rather to the scope of
**                           the frame it refers to. This is done by
**                           calling its iterate() and finished() methods
**                           from those methods called for the frame.
** 153 NVS 20071210   @36298 Moved call to finished() of aggregation
**                           handler before doing the rest of cleanup.
**                           Added missing down() call to compensate for
**                           the refactoring of iterate() method.
** 154 GES 20071213   @36327 Converted screen buffer caching to use
**                           Integer keys instead of integer (this is
**                           a bit lighter weight and allows ints to be
**                           used everywhere through autoboxing). Switch
**                           screen buffers in setScreenBuffer() if the
**                           frame-line changed during the last trip to
**                           the client.
** 155 GES 20071214   @36366 Don't swap screen buffers (setScreenBuffer)
**                           for non-down frames or when redirected, as
**                           no such swapping happens in the 4GL.
** 156 GES 20071214   @36374 Give the caller control over whether to allow
**                           screen buffer switching in setScreenBuffer.
** 157 GES 20071219   @36478 Honor dynamic down value in scroll to fix
**                           cache corruption.
** 158 SIY 20071219   @36512 Fixed displaying of totals.
**                           Ref: #36508
** 159 SIY 20071221   @36558 Fixed support of multiple accumulators in 
**                           some complex cases (conditional elements in 
**                           elements array).
**                           Ref: #36556
** 160 SIY 20071224   @36566 Do not produce accumulator results for
**                           sub-only accumulators.
**                           Ref: #36564
** 161 NVS 20071218   @36460 iterate() and retry() methods both should
**                           reset 1 down frames as well as down frames.
** 162 CA  20080104   @36670 DOWN statements in redirected mode (named or
**                           unnamed) must end any previous put output.
** 163 NVS 20080108   @36728 Only unconditional DOWNs flush any previous
**                           PUT output. Conditional DOWNs don't.
** 164 CA  20080110   @36781 Frames registered as a header inside an 
**                           iteration block must de-register themselves  
**                           from the stream's header list when the next  
**                           iteration is performed.
** 165 CA  20080115   @36814 The de-registration of a page-top/bottom frame
**                           must be done before the iterate() method is
**                           executed for that frame - it was moved to
**                           LogicalTerminal.ScopeProcessor.process. Access 
**                           modifier for some methods was changed from
**                           private to package.
** 166 CA  20080124   @36896 When a frame opens or closes a scope, this will
**                           be signaled to the client regardless of the
**                           scope nesting level. Ref: #36895
** 167 CA  20080219   @37128 When a DISPLAY ACCUM stmt accumulates its
**                           data (in AggregateHandler.accumulate), the 
**                           accumulated data must be retrieved from the
**                           AggregatorElement passed as a parameter - this
**                           is because, if a new complex expression is
**                           instantiated on each iteration, then the 
**                           accumulation would be done only with the first
**                           data (the one used to instantiate the first
**                           AggregatorElement, on the first iteration).
** 168 SIY 20080226   @37167 Editing block should handle conditional 
**                           up/down identically to regular update()/set()
**                           calls.
**                           Ref: #37166.
** 169 NVS 20080307   @37410 Added isBatch() which returns the batching
**                           mode status.
** 170 SIY 20080317   @37457 Added support for terminal stream.
**                           Ref: #37453
** 171 GES 20080317   @37591 Added FIRST-CHILD and TYPE attributes.
** 172 SIY 20080320   @37599 Added support for widgets which can have 
**                           screen values different than data value.
** 173 NVS 20080320   @37621 Changed LogicalTerminal.enable() call to
**                           match the new signature.
** 174 NVS 20080326   @37687 startEditingMode() method calls promptFor()
**                           client export since its functionality is now
**                           part of the modified promptFor().
** 175 SVL 20080402   @37787 updateDynamicTitle() call is removed from
**                           setDynamicTitle() method.
** 176 SIY 20080429   @38134 Ignore down() requests in interactive mode
**                           until frame is displayed.
**                           Ref: #38137
** 177 SIY 20080429   @38139 Preserve widget colors during setColors() in
**                           DOWN frames.
** 178 GES 20080429   @38146 Defer the push of screen definitions to do
**                           them in a batch instead of making a trip to
**                           the client for each one. This improves
**                           performance with no negative side effects.
** 179 SIY 20080430   @38155 Interactive PROMPT-FOR should set hadView 
**                           because it contains implicit VIEW.
** 180 GES 20080501   @38187 Merged endOutput calls into other client
**                           calls to avoid the extra trip.
** 181 NVS 20080507   @38231 Fixed makeFrameVisible(). Making the frame
**                           visible should follow making widgets visible
**                           or else browse, if it's in the frame, will
**                           cause pushScreenDefinition() and the frame
**                           will be made visible prematurely, which hides
**                           the frame title and may have other issues.
** 182 SIY 20080514   @38275 Push shared frame definition to client when 
**                           current incarnation finishes its life cycle.
** 183 SIY 20080604   @38546 Added accessor for frame name. Inferred 
**                           generic type arguments. 
** 184 SIY 20080609   @38648 Added accessor for hadView flag. 
** 185 NVS 20080516   @38791 Code cleanup in startEditingMode() method.
** 186 NVS 20080711   @39092 importSharedFrame() forces redrawing of
**                           dynamic title if it was assigned. Ignoring
**                           this was the root cause of the disappearing
**                           title.
** 187 SIY 20080715   @39111 Visible shared frame does not need to be 
**                           displayed again.
** 188 SIY 20080904   @39721 Inferred generics in frame creation methods.
** 189 CA  20080905   @39742 Added missing getter for VISIBLE attribute.
**                           When modifying the COL/ROW attributes of this 
**                           frame, send a flag to the client if the 
**                           location was changed.
** 190 CA  20080909   @39751 frameLineWorker will send the conditional
**                           down status to the client. This way, if the
**                           frame is hidden, FRAME-LINE will return 0; 
**                           else, the conditional down value will be 
**                           used by the client.
** 191 SIY 20080926   @39965 Track frame initialization and ignore invalid
**                           ROW/COL setting.
** 192 NVS 20081002   @39995 Fixed handling of interactive conditional
**                           DOWNs when previous output was redirected. 
** 193 CA  20080925   @39963 Support for dynamic browse title - the title
**                           expression is evaluated only the first time
**                           the buffer is sent to the client. Added new
**                           setDynamicTitle variants, which accept a BDT
**                           parameter.
** 194 CA  20080930   @39987 Added accessors for FRAME:WIDTH/HEIGHT-CHARS
**                           and FRAME-COL/ROW attributes.
** 195 CA  20081003   @40054 Fix the DISPLAY with ACCUM statement, when
**                           no accumulation is performed.
** 196 CA  20081017   @40138 The cached buffer is not sent to the client,
**                           if the frame had redirected output.
** 197 GES 20081106   @40358 On abnormal end in an editing block, don't
**                           copy changes back from the screen buffer.
** 198 SIY 20081111   @40413 Improved detection of abnormal end in editing
**                           block.
** 199 GES 20081112   @40414 Ensure that client side cleanup always gets
**                           called when an editing block exits via a
**                           path other than continueEditing().
** 200 SVL 20081107   @40392 Changes in frameLineWorker() due to API
**                           change. Changes in cache row calculation into
**                           up()/down().
** 201 CA  20081127   @40702 On scope enter, the frame must send a flag to
**                           the client, indicating that there is a parent
**                           block iterated at least once.
** 202 SIY 20081209   @40702 Reworked frame hiding implementation.
** 203 GES 20081210   @40863 Implemented a quirk of the UPDATE stmt which
**                           disables any accumulation of nested block's
**                           interactions counter. This can only happen
**                           in a trigger. Note that SET/PROMPT-FOR don't
**                           have or need this behavior.
** 204 NVS 20081211   @40874 The default field group is created right away
**                           having the frame as its parent.
**                           Fixed the NAME attribute getter for the frame
**                           as a widget.
** 205 SIY 20081226   @40995 Reset hadView flag in openScope(). 
** 206 SIY 20090122   @41194 Reset frame buffer in openScope(). 
** 207 SIY 20090127   @41231 Ignore display() calls for frame with no
**                           defined widgets.
** 208 SIY 20090206   @41266 Take into account presence of headers when  
**                           checking for empty display() call.
** 209 SIY 20090210   @41286 hadView now is set from single location (see
**                           LogicalTerminal.ScopeProcessor.frameViewed()).
**                           Added separate canHide flag which tracks
**                           frames' potential ability to hide other 
**                           frames during view() call.
** 210 SIY 20090213   @41308 display(BrowseWidget) now forces refresh.
** 211 SIY 20090220   @41341 Fixed handling of DOWN for stream redirected
**                           to interactive terminal. It should be handled
**                           just like regular interactive terminal.
** 212 CA  20090227   @41388 Added WrappedResources.isUnknown method; this
**                           implementation returns false.
** 213 ECF 20090312   @41535 Added safety code to updateDynamicTitle(). We
**                           now bracket the resolution of the title text
**                           with calls to enable/disable ErrorManager's
**                           warning mode to prevent erroring out if code
**                           invoked by this method raises an error
**                           condition.
** 214 SIY 20090316   @41615 Improved #213 in order to avoid generation of
**                           error messages during repeated frame views. 
** 215 SIY 20090407   @41730 Improved interaction of DOWN and UNDERLINE 
**                           in order to preserve correct cursor position
**                           when implicit and explicit down() calls are 
**                           mixed.
** 216 CA  20090413   @41781 Added handling for FRAME:VISIBLE/HIDDEN and 
**                           SENSITIVE flags.
** 217 CA  20090414   @41789 Added getHeaderWidgetForId, used by VISIBLE 
**                           flag change for header widgets. For shared 
**                           frames, the VISIBLE flag must be inherited
**                           from the original frame.
** 218 SIY 20090506   @42102 Honor users' attempt to override "unknown" date
**                           field.                                
** 219 SIY 20090602   @42547 Do not clear frame buffer upon retry() if 
**                           condition was observed.
** 220 GES 20090624   @42883 Protect streams access from closed streams.
** 221 SIY 20090701   @43015 Added server side part of format checking.
** 222 SIY 20090709   @43122 Fixed NPE in set().
** 223 SIY 20090716   @43204 Perform final check only for fields which were 
**                           entered during SET.
** 224 CA  20090727   @43392 When a new frame buffer is built, all static text
**                           widgets must be copied to the new frame buffer. 
** 225 SIY 20090817   @43685 Strip spaces from formatted index before using it
**                           for widget name generation.
** 226 GES 20090828   @43800 Make sure the disabling of ignore and warning
**                           modes always occurs even on exception.
** 227 SIY 20090915   @43915 Flush redirected frames in finish(), if they were  
**                           not flushed in iterate().
** 228 SIY 20090917   @43926 Retrieving values from screen buffer depends on 
**                           type of input operation. 
** 229 SIY 20091009   @44130 Changed order of checks in frameLineWorker(). 
**                           This allows significantly reduce number of 
**                           round-trips to client in certain cases.
** 230 SIY 20091117   @44403 Fixed NPE in display(FrameElement[]) for cases 
**                           when frame has no widgets and no headers (rare 
**                           case, most likely a bug in the 4GL code).
** 231 CA  20091209   @44494 The screenBufferCache must be cleared when the
**                           frame opens the first scope, as left-overs may
**                           may exist in cache.
** 233 SIY 20100316   @44749 Redirected frames should have conditional DOWN
**                           processed with their stream.
** 234 SVL 20100519   @44861 Mark the added skip as a header element.
** 235 SVL 20101004          Added support for WIDTH-PIXELS attribute.
** 236 SIY 20101004          Changes related to CONTROL-FRAME support.
** 237 SIY 20101005          Added 'UNLESS-HIDDEN' versions of enable() and
**                           display() methods (stubs).
** 238 SVL 20101006          Added support for HEIGHT, WIDTH, THREE-D, X, Y
**                           attributes and VALIDATE method.
** 239 ECF 20101006          Added asWidgetHandle method.
** 240 SIY 20101012          Added more variants for view() and hide() methods.
** 241 SIY 20101029          Changed setX/setY signatures.
** 242 SIY 20101122          Removed unused methods/variables, cleaned up 
**                           javadocs.
** 243 EVL 20101124          Implementing the approach for tracking reusage of
**                           the same internal data between file report output
**                           iteration. The objective is to fix the problems
**                           when P2J clean up the state and display value of
**                           the reusable FillIns. The change includes:
**                              - introducing the new boolean flag to store
**                                the state we want to consider
**                              - changing the flag vaue in view() method calls
**                                depending on the new incoming data presence.
**                              - bypass the screen buffer resetting in
**                                resetFrameBuf() for such cases and making
**                                cleanup in finished() method
**                              - downWorker() is modified to disable screen
**                                buffer reset and calling down() on the
**                                client side for this case
** 244 ECF 20101209          Replaced standard, J2SE dynamic proxy use with
**                           GCD proxy implementation in createFrameWorker.
**                           This required major refactoring of the code which
**                           creates standard and shared frames, and which
**                           imports shared frames, as well as the inner class
**                           SharedFrameCleaner. The new proxy implementation
**                           allows a superclass to be specified for the proxy
**                           class. This allows us to leverage many existing
**                           methods in GenericFrame, so that we do not need
**                           to re-implement these methods in every proxy.
**                           This is necessary to save PermGen memory, which
**                           otherwise would grow with each addition to the
**                           CommonFrame API, using the standard, J2SE dynamic
**                           proxy implementation. The increased PermGen for
**                           each new CommonFrame method would be multiplied
**                           by the total number of proxies created. With the
**                           new implementation, those methods are simply
**                           invoked as virtual methods in GenericFrame.
** 245 CA  20110314          Fixed SCREEN-VALUE behavior for logical widgets:
**                           the returned value must be the formatted 
**                           representation of the value currently held by the
**                           widget. Not confirmed in a 4GL environment.
** 246 ECF 20110502          Revised createFrameProxy() to proxy all of the
**                           given interface's superinterfaces, up to (but not
**                           including) CommonFrame. This is necessary for
**                           some hand-written frame code which uses a custom
**                           frame inheritance hierarchy.
** 247 SIY 20110112          Changes related to new color scheme.
** 248 SIY 20111101          Updated imports.
** 249 SIY 20111202          Updated imports.
** 250 CS  20121031          Removed client specific references
*/

package com.goldencode.p2j.ui;
 
import java.beans.*;
import java.util.*;
import java.util.logging.*;
import java.util.List;
import java.lang.reflect.*;
import com.goldencode.proxy.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.BlockManager.Condition;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;
import com.goldencode.p2j.ui.client.format.*;

/**
 * This class implements a frame factory method which produces instances of 
 * interfaces generated at conversion time.
 * <p>
 * The implementation is based on the GCD dynamic proxy library. Factory
 * generated instances are proxies. They forward method calls back to this
 * class.
 * <p>
 * The forwarded methods are:
 * <ul>
 *   <li>widget value getters;
 *   <li>widget value setters;
 *   <li>widget accessors;
 * </ul>
 * Widgets corresponds to either single variables or those defined in Progress
 * as having extents (arrays).
 * <p>
 * Arrays implement two types of methods: for single individually named 
 * elements or for generic names that require a runtime calculated index. 
 */
public class GenericFrame
implements InvocationHandler,
           CommonFrame,
           Finalizable,
           WrappedResource
{
   /** proxy getter call */
   private static final int INVOKE_GET = 0;

   /** proxy setter call */
   private static final int INVOKE_SET = 1;

   /** proxy accessor call */
   private static final int INVOKE_ACCESS = 2;

   /** Error message for enable statement during input redirection. */
   private static final String ENABLE_INPUT_REDIRECTED =
      "Cannot execute ENABLE statement when input stream is not the screen";
   
   /** logger */
   private static final Logger LOG =
      LogHelper.getLogger(GenericFrame.class.getName());
   
   /** scope nesting counter */
   private boolean scopeValid = false;
   
   /** batch modifications flag */
   private boolean batch = true;

   /** instance of the frame configuration implementation class */
   private Object configInstance = null;

   /** class of the frame configuration implementation class */
   private Class<?> configClass = null;

   /** frame name */
   private String frameName = null;

   /** frame ID */
   private int frameId = -1;

   /** Maps widget names to widget instances (from the frame definition). */
   private Map<String, GenericWidget> n2w = null;

   /** Maps converted widget names to original ones */
   private Map<String, String> n2n = null;

   /**
    * Maps header widget names to widget instances (from the frame
    * definition).
    */
   private Map<String, GenericWidget> n2h = null;
   
   /** List of widgets in ID order. */
   private GenericWidget[] widgets = null;

   /** Frame headers. */
   private FrameElement[] headers = null;
   
   /** Flag for whether a page element has been registered. */
   private boolean needsPageElementCleanup = false;
   
   /** frame config */
   private FrameWidget frame = null;

   /** frame definition */
   private ScreenDefinition frameDef = null;
      
   /** screen buffer for this frame */
   private ScreenBuffer frameBuf = null;
   
   /** Identifies type of editing block (PROMPT_FOR, SET or UPDATE). */
   private int currentOp = NONE;
   
   /** Lock object for synchronization of shared editing block data. */
   private Object editLock = new Object();

   /** The widget used for displaying a statistic's description. */
   private GenericWidget statsColumn = null;
   
   /** Stream to which redirected aggregator element output is sent */
   private Stream redirOut = null;
   
   /** The aggregate handler currently active for this frame */
   private AggregateHandler aggregateHandler = null;
   
   /**
    * Keeps state of the conditional DOWN and/or UP cursor movements.
    * Can be set to {-1, 0, +1}. -1 means there is an outstanding request
    * to move the cursor up. 0 means no outstanding requests. +1 means there
    * is an outstanding request to move the cursor down.
    */
   private int pendingUpDown = 0;
   
   /** Callback expression for the number of iterations in a down frame. */
   private Resolvable dynDown = null;

   /** Callback expression for the frame column. */
   private Resolvable dynColumn = null;

   /** Callback expression for the frame row. */
   private Resolvable dynRow = null;

   /** The registered storage for the current frame title. */
   private CharacterExpression dynTitle = null;

   /** Set to true when frame is viewed first time. */
   private boolean hadView = false;

   /** List of widgets in text groups in current editing block. */
   private List<FrameElement> textGroup = null;

   /** Nesting scope counter. */
   private int scopeNesting = 0;

   /** Shared flame flag. */
   private boolean shared = false;

   /** A screen-buffer cache to be used with a down frame. 
    * key   - a row number
    * value - a ScreenBuffer instance
    */
   private Map<Integer, ScreenBuffer> screenBufferCache = 
                                    new HashMap<Integer, ScreenBuffer>();

   /** Dynamic DOWN value. */
   private int dynamicDown = 0;
   
   /** The first field group in the frame (holds all non-header widgets). */
   private FieldGroup fieldGroup = null;

   /**
    * Set to <code>true</code> when frame is initialized by appropriate
    * {@link Settable} instance.
    */
   private boolean insideSetup = false;

   /** 
    * List of aggregators, deferred to be registered on the first iteration. 
    */
   private FrameElement[] deferredAggregators = null;

   /** Can hide other frames. */
   private boolean canHide = true;

   /** Set to <code>true</code> if last operation was DOWN. */
   private boolean afterDown = false;

   /** Set to <code>true</code> if last down() was internal underline() call. */
   private boolean afterUnderline = false;

   /** Track presence of condition during frame processing. */
   private boolean hadCondition = false;

   /** Track implicit validation nesting. */
   private int triggerActive = 0;

   /** Final format check control. */
   private boolean finalCheck = false;

   /** Second and subsequent formatting errors are suppressed. */
   private boolean silentFormatErrors = false;

   /** Widget IDs stored in this map are not checked for formatting errors. */
   private Set<Integer> formatFilter = new HashSet<Integer>();

   /** Track frame data flushing in redirected mode. */
   private boolean flushed = true;
   
   /** Track whether we change data from one iteration to another */
   private boolean sameViewDataRedirected = false;

   /**
    * Default constructor; creates a new instance of a frame. This constructor
    * is meant to be invoked by subclasses. To initialize the frame, the
    * caller must invoke {@link #initialize(Class, Object, String)}.
    */
   protected GenericFrame()
   {
   }
   
   /**
    * Creates new frame as a proxy object (as a local resource).  This loads
    * the static frame definition.
    *
    * @param    <T> 
    *           Frame interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    *
    * @return   The created frame.
    */
   public static <T extends CommonFrame> T
   createFrame(Class<T> iface, String name)
   {
      T frame = createFrameProxy(iface);
      initializeNewFrame((GenericFrame) frame, iface, name);
      setupFrame(frame);
      
      return frame;
   }

   /**
    * Creates new frame as a proxy object and exports that new frame as a
    * shared resource (non-global).  This loads the static frame definition.
    *
    * @param    <T> 
    *           Frame interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    *
    * @return   The created frame.
    */
   public static <T extends CommonFrame> T
   createSharedFrame(Class<T> iface, String name)
   {
      T frame = createFrame(iface, name);
      
      GenericFrame gFrame = (GenericFrame) frame;
      gFrame.shared = true;
      
      // export the frame (there is no global shared frame)
      GenericFrame old = SharedVariableManager.addFrame(name, gFrame);
      
      if (old != null)
         LogicalTerminal.releaseSharedFrame(old.getFrameId());
      
      // return the specific proxy that matches the requested interface
      return frame;
   }

   /**
    * Import an existing frame as a proxy object from the shared resource
    * manager.  Create a new proxy and copy the state of the retrieved, shared
    * frame to this new proxy for temporary use, until the scope in which the
    * new proxy is used ends.  When the associated external procedure exits,
    * the original frame definition will be restored, and any relevant state
    * from the exiting scope will be copied back to it.
    * 
    * @param    <T> 
    *           Frame interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    *
    * @return   The imported frame.
    */
   public static <T extends CommonFrame> T importSharedFrame(Class<T> iface,
                                                             String name)
   {
      T proxy = createFrameProxy(iface);
      GenericFrame newFrame = (GenericFrame) proxy;
      
      Object cfgInst = getConfigInstance(iface);
      Class<?> cfgCls = cfgInst.getClass();
      
      // lookup the old frame from the shared resource manager
      GenericFrame oldFrame = (GenericFrame)
         SharedVariableManager.lookupFrame(name);
      
      LogicalTerminal.activateSharedFrame(oldFrame.getFrameId());
      
      // check current visibility of old frame and store into new frame
      boolean visible = oldFrame.frame.isVisible().booleanValue();

      // save visibility state for each widget
      Map<Integer, Boolean> wVis = new HashMap<Integer, Boolean>();
      for (int i = 0; i < oldFrame.widgets.length; i++)
      {
         GenericWidget w = oldFrame.widgets[i];
         wVis.put(w.getId(), w.config.isVisible());
      }
      
      // save visibility state for header widgets too
      Iterator<GenericWidget> itr = oldFrame.n2h.values().iterator();
      while (itr.hasNext())
      {
         GenericWidget w = itr.next();
         wVis.put(w.getId(), w.config.isVisible());
      }
      
      // the following object copies the state of the old frame into the new
      // frame at construction;  the new frame will be used in the newer scope
      // and selected state (which may have been updated in the newer scope)
      // will be copied back to the old frame when the new scope ends
      Finalizable clean = new SharedFrameCleaner(oldFrame, newFrame);
      TransactionManager.registerNextExternal(clean);
      
      // honor any overriding definition differences for this instance of
      // the shared frame
      newFrame.coreInitialize(cfgCls, cfgInst, name);
      
      newFrame.frameDef = null;
      newFrame.batch    = true;
      
      // restore visible state for frame and widgets into new frame
      newFrame.frame.setVisible(visible);
      Iterator<Integer> itrw = wVis.keySet().iterator();
      while (itrw.hasNext())
      {
         int wid = itrw.next();
         visible = wVis.get(wid);
         
         GenericWidget w = newFrame.getWidgetForId(wid);
         if (w == null)
         {
            // if not found, it must be a header widget
            w = newFrame.getHeaderWidgetForId(wid);
         }
         if (w != null)
         {
            w.config.setVisible(visible);
         }
      }
      
      // the newly recreated frame should honor the dynamic title if any
      // the call below takes care of this, conditionally
      newFrame.updateDynamicTitle();
      
      // execute frame-specific setup
      setupFrame(proxy);

      return proxy;
   }
   
   /**
    * Retrieve the frame definition class and create and return a new instance
    * of this class.  This method relies on the convention of storing the
    * frame definition class in a static field of the frame interface, named
    * "configClass".  Reflection is used to retrieve that field and
    * instantiate the object from it.
    * 
    * @param   iface
    *          Frame interface type.
    * 
    * @return  Instance of the frame definition class.
    * 
    * @throws  RuntimeException
    *          if the configClass field does not exist, is inaccessible, or
    *          the class cannot be instantiated.
    */
   private static Object getConfigInstance(Class<?> iface)
   {
      Class<?> cfgCls  = null;
      Object cfgInst = null;

      try
      {
         // each frame interface has a static field "configClass", which
         // contains the implementation class associated with that interface
         Field link = iface.getDeclaredField("configClass");
         
         cfgCls  = (Class<?>) link.get(null);
         cfgInst = cfgCls.newInstance();
      }
      catch (NoSuchFieldException exc)
      {
         throw new RuntimeException("configClass is undefined ", exc);
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException("configClass is inaccessible ", exc);
      }
      catch (InstantiationException exc)
      {
         throw new RuntimeException("cannot instantiate configClass ", exc);
      }
      
      return cfgInst;
   }
   
   /**
    * Initialize the frame and register it with the transaction manager.
    *
    * @param <T>
    *
    * @param    frame
    *           Frame to be initialized (also the proxy for the given
    *           interface).
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    */
   private static <T extends CommonFrame>
   void initializeNewFrame(GenericFrame frame,
                           Class<T> iface,
                           String name)
   {
      Object configInstance = getConfigInstance(iface);
      Class<?> impl = configInstance.getClass();
      
      // initialize the frame (aka, invocation handler) for the implementation
      // class
      frame.initialize(impl, configInstance, name);

      TransactionManager.registerNextExternal(new Callback(frame));
   }

   /**
    * Create and initialize a frame as a proxy object, loading the static
    * frame definition.
    * 
    * @param    <T> 
    *           Frame interface.
    * @param    iface
    *           The class of the interface definition. 
    *
    * @return   The created frame.
    */
   @SuppressWarnings("unchecked")
   private static <T extends CommonFrame> T createFrameProxy(Class<T> iface)
   {
      List<Class<?>> ifaceList = new ArrayList<Class<?>>();
      
      // proxy the interface and all of its ancestor superinterfaces, except
      // CommonFrame
      Class<?> next = iface;
      while (next != null && !CommonFrame.class.equals(next))
      {
         ifaceList.add(next);
         Class<?>[] ext = next.getInterfaces();
         next = (ext.length == 1 ? ext[0] : null);
      }
      
      // also proxy Settable
      ifaceList.add(Settable.class);
      
      Class<?>[] ifaces = new Class<?>[ifaceList.size()];
      ifaces = ifaceList.toArray(ifaces);
      
      // manufacture an instance that will redirect method calls to our
      // invocation handler
      T proxy = (T) ProxyFactory.getProxy(GenericFrame.class, ifaces);
      
      return proxy;
   }
   
   /**
    * Invoke the frame's setup logic and set the batch flag to false.
    * 
    * @param   <T>
    *          Frame interface type.
    * @param   frame
    *          Frame proxy instance.
    */
   private static <T extends CommonFrame> void setupFrame(T frame)
   {
      frame.setInsideSetup(true);
      ((Settable) frame).setup(frame);
      frame.setInsideSetup(false);
      frame.batch(false);
   }
   
   /**
    * @see com.goldencode.p2j.ui.CommonFrame#setInsideSetup(boolean)
    */
   @Override
   public void setInsideSetup(boolean flag)
   {
      insideSetup = flag;
   }
   
   /**
    * Get frame name.
    * 
    * @return frame name.
    */
   public String getName()
   {
      return frameName;
   }
   
   /**
    * Check if frame is shared.
    * 
    * @return  <code>true</code> if frame is shared frame.
    */
   public boolean isShared()
   {
      return shared;
   }

   /**
    * Getter for the redirOut field.
    * 
    * @return  the redirOut field.
    */
   public Stream getRedirOut() 
   {
      return redirOut;
   }
   
   /**
    * Register header frame elements.
    * <p>
    * This method corresponds to these Progress statements:
    * FORM HEADER
    * 
    * @param   fe
    *          Array of headers to register.
    */
   @Override
   public void registerHeader(FrameElement[] fe)
   {
      headers = fe;
   }

   /**
    * Register all the widgets which need accumulation with this frame.
    * 
    * @param    fe
    *           The list of widgets that may need accumulation.
    */
   @Override
   public void registerAccumulators(FrameElement[] fe)
   {
      this.deferredAggregators = fe;
   }
   
   /**
    * This method acts as an interceptor which is called every time a
    * method of the proxy object is invoked. It allows us to perform pre-
    * and post-processing for these methods, and to delegate the regular
    * work of the method to the backing implementation DMO.
    *
    * @param   proxy
    *          Proxy object upon which the intercepted method is invoked.
    * @param   method
    *          Method to be invoked through reflection.
    * @param   args
    *          Arguments, if any, to <code>method</code>.
    *
    * @return  value returned from the invoked method. 
    * 
    * @throws  Throwable
    *          if an exception/error occurs upon the invocation target,
    *          or if any problem occurs during pre- or post-processing.
    */
   @Override
   public Object invoke(Object proxy, Method method, Object[] args)
   throws Throwable
   {
      int    op         = -1;
      String widgetName = null;

      try
      {
         Class<?> declClass = method.getDeclaringClass();
         
         // handle special (reserved) methods
         if (declClass.equals(Object.class)      ||
             declClass.equals(Settable.class))
         {
            return method.invoke(configInstance, args);
         }
         
         // handle CommonFrame interface methods
         if (declClass.equals(CommonFrame.class))
         {
            return method.invoke(this, args);
         }
         
         int remove = -1;
         
         String wname = method.getName();
         
         // must be a getter/setter/widget accessor
         if (wname.startsWith("get"))
         {
            // standard getter
            op = INVOKE_GET;
            remove = 3;
         }
         else if (wname.startsWith("is"))
         {
            // boolean getter
            op = INVOKE_GET;
            remove = 2;
         }
         else if (wname.startsWith("set"))
         {
            // setter
            op = INVOKE_SET;
            remove = 3;
         }
         else if (wname.startsWith("widget"))
         {
            // widget access
            op = INVOKE_ACCESS;
            remove = 6;
         }
         else
         {
            throw new RuntimeException("unknown proxy method call " + wname);         
         }
         
         widgetName = wname.substring(remove);
      }
      
      catch (InvocationTargetException exc)
      {
         // This is the result of a programming error.  Don't wrap in a
         // legacy-style exception type.
         throw exc.getTargetException();
      }

      GenericWidget widget = getWidget(widgetName,
                                       op, 
                                       (args == null) ? 0    : args.length,
                                       (args == null) ? null : args[0]);
      
      
      // execute getter
      if (op == INVOKE_GET)
      {
         return getter(getWidgetId(widget), method.getReturnType(), true, true);
      }

      // execute setter
      if (op == INVOKE_SET)
      {
         Object parm = (args != null && args.length == 2) ? args[1] : args[0];
         return setter(getWidgetId(widget), parm);
      }

      // must be the accessor
      
      if (op == INVOKE_ACCESS && widget == null)
      {
         if (args != null && args.length == 1)
         {
            String err = "Array index " + indexAsString(args[0]) +
                         " is not formatted in the frame";
            ErrorManager.recordOrThrowError(622, err);
            
            // silent error case
            return null;
         }
      }
      
      return widget;
   }
   

   /**
    * Sets the VISIBLE attribute.
    *
    * @param    visible
    *           <code>true</code> if the attribute must be set to ON.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible) 
         view();
      else 
         hide();
   }

   /**
    * Sets the VISIBLE attribute.
    *
    * @param    visible
    *           <code>true</code> if the attribute must be set to ON.
    */
   @Override
   public void setVisible(logical visible)
   {
      if (visible.isUnknown())
         return;
      setVisible(visible.booleanValue());
   }

   /**
    * Gets the VISIBLE attribute of the frame.
    *
    * @return   <code>true</code> if the attribute is set to ON.
    */
   public boolean _isVisible()
   {
      return LogicalTerminal.isVisible(frameId);
   }

   /**
    * Gets the VISIBLE attribute of the frame.
    *
    * @return   PROGRESS <code>true</code> value if the attribute is set to ON.
    */
   @Override
   public logical isVisible()
   {
      return new logical(_isVisible());
   }

   /**
    * Queries the SENSITIVE attribute of this frame.
    *
    * @return <code>true</code> if SENSITIVE is set.
    */
   @Override
   public logical isSensitive()
   {
      return new logical(frame.isSensitive());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(logical s)
   {
      if (s.isUnknown())
         return;
      setSensitive(s.booleanValue());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(boolean s)
   {
      frame.setSensitive(s);
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @Override
   public void setHidden(logical hidden) 
   {
      if (hidden.isUnknown())
         return;
      setHidden(hidden.booleanValue());
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @Override
   public void setHidden(boolean hidden) 
   {
      setVisible(!hidden);
      frame.setHidden(hidden);
   }

   /**
    * Queries the HIDDEN attribute of this frame.
    *
    * @return <code>true</code> if HIDDEN is set.
    */
   @Override
   public logical isHidden()
   {
      return new logical(frame.isHidden());
   }
   
   /**
    * Obtain the current value in the screen buffer for the given widget and 
    * assign this into the backing variable. If there is no value in the
    * screen buffer (if it is <code>null</code>) no assignment will occur.
    *
    * @param    data
    *           Represents the variable to assign and the widget for which
    *           the screen value is to be retrieved.
    */
   @Override
   public void assignScreenValue(FrameElement data)
   {
      GenericWidget widget = data.getWidget();
      int           wid    = getWidgetId(widget);
      Class<?>      type   = ((ControlEntity) widget).getDataClass();
      
      // get the value out of the screen buffer (and make sure it is 
      // converted to the right type if needed)
      BaseDataType value = (BaseDataType) getter(wid, type, false, false);
         
      // only assign if there is a value in the screen buffer
      if (value != null)
      {
         data.set(value);
      }
   }
   
   /**
    * Get this frame as instance of <code>GenericFrame</code>.
    * 
    * @return  reference to this frame as <code>GenericFrame</code>
    *          instance.
    */
   @Override
   public FrameWidget asWidget()
   {
      return frame;
   }
   
   /**
    * Get this frame wrapped in a handle.
    * 
    * @return  handle to this frame.
    */
   @Override
   public handle asWidgetHandle()
   {
      return new handle(frame);
   }
   
   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param event
    *        event to be applied
    */
   @Override
   public void apply(character event)
   {
      if (event.isUnknown())
         throw new RuntimeException("event cannot be unknown");

      apply(event.getValue());
   }

   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param event
    *        event to be applied
    */
   @Override
   public void apply(String event)
   {
      LogicalTerminal.apply(this, event);
   }

   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param eventCode
    *        event to be applied
    */
   @Override
   public void apply(integer eventCode)
   {
      if (eventCode.isUnknown())
         throw new RuntimeException("event cannot be unknown");

      LogicalTerminal.apply(this, eventCode.intValue());
   }

   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param eventCode
    *        event to be applied
    */
   @Override
   public void apply(double eventCode)
   {
      LogicalTerminal.apply(this, (int)eventCode);
   }

   /**
    * Sets batching mode for frame modification.
    *
    * @param  on
    *         <code>true</code> sets batching mode
    */
   @Override
   public void batch(boolean on)
   {
      batch(on, false);
   }

   /**
    * Queries batching mode for frame modification.
    *
    * @return 
    *         <code>true</code> if batching mode is in effect
    */
   public boolean isBatch()
   {
      return batch;
   }

   /**
    * Sets batching mode for frame modification and optionally pushes the
    * screen definition to the client. 
    *
    * @param  on
    *         <code>true</code> sets batching mode
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push. This is only honored when <code>on</code> is
    *         <code>false</code>.
    */
   public void batch(boolean on, boolean force)
   {
      batch = on;
      
      if (!batch)
         pushScreenDefinition(force);
   }
   
   /**
    * Performs CHOOSE processing.
    * <p>
    * This method corresponds to these Progress statements:
    * CHOOSE
    *
    * @param  c
    *         CHOOSE statement options
    */
   @Override
   public void choose(Choose c)
   {
      conditionalUpDown();
      LogicalTerminal.choose(this, c, null);
   }
      
   /**
    * Performs CHOOSE processing.
    * <p>
    * This method corresponds to these Progress statements:
    * CHOOSE COLOR
    *
    * @param  c
    *         CHOOSE statement options
    * @param  cs
    *         color specification for use with CHOOSE
    */
   @Override
   public void choose(Choose c, ColorSpec cs)
   {
      conditionalUpDown();
      LogicalTerminal.choose(this, c, cs);
   }
      
   /**
    * Clears the data and colors for all fill-in fields of the frame.
    * For down frames, also clears side labels.
    * <p>
    * This method corresponds to these Progress statements:
    * CLEAR
    *
    * @param  noPause
    *         <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void clear(boolean noPause)
   {
      conditionalUpDown();
      resetFrameBuf();
      LogicalTerminal.clear(this, noPause);
   }

   /**
    * Clears all occurrences for the down frame and resets the current display
    * position to the top of the frame.
    * <p>
    * This method corresponds to these Progress statements:
    * CLEAR ALL
    *
    * @param  noPause
    *         <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void clearAll(boolean noPause)
   {
      pendingUpDown = 0;
      resetFrameBuf();
      LogicalTerminal.clearAll(this, noPause);
      
      // the screen buffer cache needs also to be cleared
      screenBufferCache.clear();
   }
      
   /**
    * Disables input for all field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements:
    * DISABLE ALL
    */
   @Override
   public void disable()
   {
      LogicalTerminal.disable(this, null);
   }

   /**
    * Disables input for all field-level widgets of the frame, except the
    * specified ones.
    * <p>
    * This method corresponds to these Progress statements:
    * DISABLE ALL EXCEPT
    *
    * @param    xcpt
    *           Array ow widgets to be excluded from the operation.
    */
   @Override
   public void disableExcept(GenericWidget[] xcpt)
   {
      int[] ids = remainingIds(xcpt);
      
      if (ids == null)
      {
         // all widgets are excluded
         return;
      }
      
      // disable remaining widgets
      LogicalTerminal.disable(this, ids);
   }

   /**
    * Disables input for all field-level widgets of the frame, except the
    * specified one.
    * <p>
    * This method corresponds to these Progress statements:
    * DISABLE ALL EXCEPT
    *
    * @param    widget
    *           Widgets which will be excluded from the operation.
    */
   @Override
   public void disableExcept(GenericWidget widget)
   {
      disableExcept(new GenericWidget[] { widget });
   }
   
   /**
    * Disables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: DISABLE
    * 
    * @param    list
    *           The widgets to disable.
    */
   @Override
   public void disable(GenericWidget[] list)
   {
      int[] ids = validIds(list);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      // disable widgets
      LogicalTerminal.disable(this, ids);
   }
   
   /**
    * Disables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: DISABLE
    * 
    * @param   list
    *          The widgets to disable.
    */
   @Override
   public void disable(FrameElement[] list)
   {
      GenericWidget[] wids = widgetListFromFrameElements(list);
      disable(wids);
   }
   
   /**
    * Displays a frame with no "data container" fields on the screen.
    * <p>
    * As part of the processing, the frame marked as visible.  The frame will
    * be brought to the top of the Z-order and refreshed.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    */
   @Override
   public void display()
   {
      display(new FrameElement[]{});
   }

   /**
    * Refresh specified browse widget.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY WITH BROWSE
    *
    * @param    browse
    *           A widget to refresh. 
    */
   @Override
   public void display(BrowseWidget browse)
   {
      browse.forceRefresh();
   }
   
   /**
    * Displays (writes) non-data-containing fields from the frame to the
    * given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void display(Stream out)   
   {
      viewWorker(out, null, false);
   }

   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void display(FrameElement[] data)
   {
      // ignore call if there are no widgets defined in frame
      if (data != null        && 
          data.length == 0    && 
          (widgets == null || widgets.length == 0) &&
          (headers == null || headers.length == 0))
      {
         return;
      }

      conditionalUpDown();
      
      Stream out = UnnamedStreams.output(); 
      
      if (out != null)
      {
         display(out, data);
      }
      else
      {
         // set the value in the screen buffer for each widget using the widget
         // id as the index
         copyToScreenBuffer(data);
      
         // now make the frame visible, bring it to the top and refresh the
         // contents
         view(data);
      }
      
      // handle aggregation if needed
      processAggregates(data, true);
   }

   /**
    * Displays (writes) a list of fields from the frame on the given stream.
    * As part of this operation, the list of data values are copied to the
    * screen buffer.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void display(Stream out, FrameElement[] data)
   {
      // set the value in the screen buffer for each widget using the widget
      // id as the index
      copyToScreenBuffer(data);
      
      viewWorker(out, data, false);

      // handle aggregation if needed
      processAggregates(data, true);

      // the registration of the frame as a page element was done in 
      // ThinClient.view_ - so the only thing left to do here is to set the 
      // needsPageElementCleanup flag to true, so the finished() method will
      // remove the frame from the stream's page-element list
      if (isPageElement(out))
      {
         needsPageElementCleanup = true;
      }
   }

   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY UNLESS-HIDDEN
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void displayUnlessHidden(FrameElement[] data)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to enable.
    */
   @Override
   public void enableUnlessHidden(GenericWidget[] list)
   {
      throw new UnsupportedOperationException("Not implemented");
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to enable.
    */
   @Override
   public void enableUnlessHidden(FrameElement[] list)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Positions the cursor on a new line in a down frame. This is a
    * conditional call. Multiple conditional down() calls have the effect of a
    * single down(1) call, which is delayed till the nearest terminal I/O
    * operation and can be cancelled by a conditional up() call.
    * <p>
    * Note that this call can be ignored in interactive mode if frame is not
    * yet viewed at least once before this call. After first VIEW
    * (explicit or implicit) hiding frame (either via HIDE statement or by
    * changing VISIBLE attribute) does not affect behavior and calls to this
    * method are honored. This corresponds to 4GL behavior.
    * <p>
    * This method corresponds to these Progress statements: DOWN
    */
   @Override
   public void down()
   {
      if (aggregateHandler != null)
         aggregateHandler.notifyDown(0);
      
      flushed  = true;
      
      // If either the unnamed stream is redirected or the last used stream
      // was something else, this call has to be passed down to the client
      // to switch streams properly.
      // Note that we assume that TERMINAL stream behaves just like interactive
      // mode and does not process DOWN immediately.
      if (hadView && isRedirected() || (isOutRedirected() && !isOutTerm()))
      {
         pendingUpDown = 0;

         if (isRedirected())
         {
            // the unnamed stream is redirected
            downWorker(1,           // down for 1 line
                       true);       // conditional down
         }
         else
         {
            // the last used stream was something else, needs switching
            downWorker(1,           // down for 1 line
                       -1,          // unnamed stream
                       true,        // conditional down
                       false);      // does not end output (?)
         }

         return;
      }

      if (hadView && pendingUpDown < 1)
      {
         int frameLine =
            LogicalTerminal.getFrameLine(this, pendingUpDown).intValue();
         
         pendingUpDown++;

         // non-down frames maintain the pending value but don't swap buffers
         if (!frame.isDown())
            return;
         
         // we add the screen buffer for the previous frame line
         addScreenBufferToCache(this.frameBuf, frameLine);

         // restore the buffer for the next row, but don't send it to the 
         // client
         int nextLine = frameLine + 1;
         frameBuf = getScreenBufferFromCache(nextLine);
      }
   }

   /**
    * Positions the cursor on a new line in a down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN n
    *
    * @param lines
    *        number of iterations of the data to move down
    */
   @Override
   public void down(double lines)
   {
      downWorker((int) lines, false);
   }

   /**
    * Positions the cursor on a new line in a down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN n
    *
    * @param lines
    *        number of iterations of the data to move down
    */
   @Override
   public void down(NumberType lines)
   {
      downWorker(lines.intValue(), false);
   }

   /**
    * Writes a single new-line to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN STREAM s
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void down(Stream out)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      downWorker(1, getRemoteStreamId(out), true, false);
   }

   /**
    * Writes <code>lines</code> number of new-lines to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           Number of new-lines to output to the stream.
    */
   @Override
   public void down(Stream out, double lines)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      if (lines < 1)
         return;
      
      downWorker((int) lines, getRemoteStreamId(out), false, true);
   }

   /**
    * Writes <code>lines</code> number of new-lines to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           Number of new-lines to output to the stream.
    */
   @Override
   public void down(Stream out, NumberType lines)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      int l = lines.intValue();

      if (l < 1)
         return;

      downWorker(l, getRemoteStreamId(out), false, true);
   }

   /**
    * Enables input for all field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements:
    * ENABLE ALL
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enable()
   {
      if (errorOnRedirectedInput(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in silent error mode
         return;
      }
          
      makeFrameVisible();
      LogicalTerminal.enable(this, null);
   }

   /**
    * Enables input for all field-level widgets of the frame, except the
    * specified ones.
    * <p>
    * This method corresponds to these Progress statements:
    * ENABLE ALL
    *
    * @param    xcpt
    *           Array ow widgets to be excluded from the operation.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enableExcept(GenericWidget[] xcpt)
   {
      if (errorOnRedirectedInput(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in silent error mode
         return;
      }
          
      int[] ids = remainingIds(xcpt);
      
      if (ids == null)
      {
         // all widgets are excluded
         return;
      }
      
      // enable remaining widgets
      LogicalTerminal.enable(this, ids);
   }

   /**
    * Enables input for all field-level widgets of the frame, except the
    * specified one.
    * <p>
    * This method corresponds to these Progress statements:
    * ENABLE ALL
    *
    * @param    widget
    *           Widget which will be excluded from the operation.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enableExcept(GenericWidget widget)
   {
      enableExcept(new GenericWidget[] { widget });
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: ENABLE
    * 
    * @param    list
    *           The widgets to enable.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enable(GenericWidget[] list)
   {
      if (errorOnRedirectedInput(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in silent error mode
         return;
      }
          
      int[] ids = validIds(list);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      // enable widgets
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] instanceof BrowseWidget)
            ((BrowseWidget) list[i]).finishSetup();
      }

      LogicalTerminal.enable(this, ids);
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: ENABLE
    * 
    * @param   list
    *          The widgets to enable.
    * 
    * @throws  ErrorConditionException
    *          If the unnamed input is redirected and silent error mode is
    *          off.
    */
   @Override
   public void enable(FrameElement[] list)
   {
      GenericWidget[] wids = widgetListFromFrameElements(list);
      enable(wids);
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * the given widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @param    widget
    *           The widget for which the value is to be retrieved.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @Override
   public character getScreenValue(GenericWidget widget)
   {
      BaseDataType value;
      
      value = (BaseDataType) convertScreenValue(character.class, 
                                      frameBuf.getScreenValue(widget.getId()),
                                      widget.getFormat());

      if (value == null)
         value = (BaseDataType) getter(widget.getId(),
                                       character.class,
                                       true,
                                       false);
      
      // the getter is only guaranteed to return a character value when
      // the value is uninitialized, otherwise we get back whatever type 
      // the data actually has
      if (!(value instanceof character))
      {
         // convert it to character
         value = character.valueOf(value);
      }
      
      return (character) value;
   }
   
   /**
    * Set the current value in the screen buffer of the backing data for
    * the given widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    widget
    *           The widget for which the value is to be retrieved.
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   @Override
   public void setScreenValue(GenericWidget widget, character value)
   {
      int wid = widget.getId();
      
      if (value == null)
      {
         // force the value to be uninitialized
         frameBuf.setState(wid, ScreenBuffer.UNINITIALIZED);
      }
      else
      {
         // make a copy
         BaseDataType val = value.duplicate();
            
         // honor the type of the widget's data
         Class<?> type = ((ControlEntity) widget).getDataClass();         
            
         // convert the data to the right type if needed
         if (!type.equals(character.class))
         {
            if (type.equals(date.class))
            {
               val = new date(value);
            }
            else if (type.equals(decimal.class))
            {
               val = new decimal(value);
            }
            else if (type.equals(integer.class))
            {
               val = new integer(value);
            }
            else if (type.equals(logical.class))
            {
               val = new logical(value);
            }
         }
            
         // set the value in the screen buffer by index
         frameBuf.putWidgetValue(wid, val);

         conditionalUpDown();
         
         // refresh the screen
         LogicalTerminal.refreshFrameWidget(this, wid, val);
      }      
   }
   
   /**
    * Sets the containing frame instance.
    * 
    * @param    hFrame
    *           handle which is expected to contain a <code>FrameWidget</code>
    *           instance
    */
   @Override
   public void setFrame(handle hFrame)
   {
      // TODO:  no idea if this works...
      this.frame = ((FrameWidget) hFrame.get());
   }

   /**
    * Set the current value in the screen buffer of the backing data for
    * the given widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    wid
    *           The widget for which the value is to be set
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   public void setFrameValue(int wid, BaseDataType value)
   {
      GenericWidget widget = getWidgetForId(wid);
      
      if (value != null && !(value instanceof character))
      {
         // honor the type of the widget's data
         Class<?> type = ((ControlEntity) widget).getDataClass();         
            
         // ensure the right type is being used
         if (!type.equals(value.getClass()))
         {
            String err = "Incompatible data types for frame-value assignment";
            ErrorManager.recordOrThrowError(676, err);
            return;
         }
      }
         
      // make a copy
      BaseDataType val = value.duplicate();
         
      if (value instanceof unknown)
      {
         val = translateUnknown(value, (ControlEntity) widget);
      }
      
      // set the value in the screen buffer by index
      frameBuf.putWidgetValue(wid, val);
      
      conditionalUpDown();
      
      // refresh the screen and own copy of frame-value
      LogicalTerminal.refreshFrameWidget(this, wid, val);
   }
   
   /**
    * Makes the frame invisible with the default pause.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE FRAME
    */
   @Override
   public void hide()
   {
      hide(false);
   }

   /**
    * Makes the frame invisible.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE FRAME
    *
    * @param  noPause
    *         <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(boolean noPause)
   {
      Stream out = UnnamedStreams.output();
      
      if (out != null)
      {
         hide(out, noPause);
      }
      else
      {
         ((FrameConfig)frame.getConfig()).setVisible(false);
         LogicalTerminal.hide(frameId, noPause);
         
         // reset the title so next VIEW/DISPLAY will start from scratch
         if (dynTitle != null)
            frameBuf.putTitle(null);
      }
   }

   /**
    * If the frame is a PAGE-TOP or PAGE-BOTTOM and the stream is paged,
    * the frame is removed from the list of those used for the header or
    * footer.  Otherwise this is a no-operation.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(Stream out, boolean noPause)
   {
      if (out != null)
      {
         // the version of HIDE that takes a widget list has a strange case
         // where the error messages are doubled (must be something deep inside
         // the Progress UI implementation) which occurs whenever there is > 1
         // widget in the list; since this method is the "all frame" version,
         // the same problem doesn't occur here; to implement that quirk please
         // see viewWorker(Stream...) since VIEW/DISPLAY have a similar issue
         
         // normal safety code here, exit in NO-ERROR mode if needed
         if (out.protectWrites())
            return;
      }
      
      // check if we must deregister our page element
      if (isPageElement(out))
      {
         // deregister the page element
         LogicalTerminal.deregisterPageElement(frameId,
                                               getPageTop(),
                                               getRemoteStreamId(out));
      }
   }
   
   /**
    * Specifies the field to be put in focus during the next input operation.
    * <p>
    * This method corresponds to these Progress statements:
    * NEXT-PROMPT
    *
    * @param  field
    *         widget to be put in focus
    */
   @Override
   public void nextPrompt(GenericWidget field) 
   {
      int id = getWidgetId(field);
      
      if (id != -1)
      {
         LogicalTerminal.nextPrompt(this, id);
      }
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    */
   @Override
   public void promptFor(FrameElement[] widgets)
   {
      Stream in = UnnamedStreams.input();
      
      setStaticWidgets(widgets);
      
      if (in != null)
      {
         // redirected input case
         promptFor(in, widgets);
      }
      else
      {
         // interactive case
         promptFor(widgets, null);
      }
   }
   
   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. Data
    * will not be copied to or from the user's variables.
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    widgets
    *           The list of widgets which are to receive input from the
    *           stream.
    */
   @Override
   public void promptFor(Stream in, FrameElement[] widgets)
   {
      processStreamInput(in, widgets, false);
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void promptFor(FrameElement[] widgets, EventList wait)   
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         promptFor(in, widgets);
      }
      else
      {
         setStaticWidgets(widgets);
         
         // process text grouping (word wrap mode)
         List<FrameElement> list = textGroup(widgets);
         
         // interactive case
         GenericWidget[] wids = widgetListFromFrameElements(widgets);
         promptFor(wids, wait);
         
         // reset text grouping (if any)
         textUngroup(list);
      }
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    */
   @Override
   public void promptFor(GenericWidget[] widgets)
   {
      promptFor(widgets, null);
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void promptFor(GenericWidget[] widgets, EventList wait)
   {
      int[] ids      = validIds(widgets);
      int   tempTerm = -1;
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      conditionalUpDown();

      // special "double-vision" mode (no input redirection but output is
      // redirected, so we must allow interactive editing but then ensure
      // that the output is echoed to the redirected output stream
      if (UnnamedStreams.output() != null)
      {
         tempTerm = LogicalTerminal.switchOutput(-1, false);
      }
      
      Set<Integer> save = formatFilter;
      
      try
      {
         if (wait == null)
         {
            wait = new EventList(true);
         }
         wait.setOperation(EventList.PROMPT_FOR);
         
         formatFilter = new HashSet<Integer>();
         
         for (int i = 0; i < ids.length; i++)
         {
            Object value = frameBuf.getWidgetValue(ids[i]);
            
            if (value instanceof character)
            {
               character chr = (character) value;
               
               if (!chr.isUnknown())
               {
                  if (chr.toStringMessage().trim().isEmpty())
                     formatFilter.add(ids[i]);
                  else
                     enableFinalFormatCheck();
               }
            }
         }
         
         LogicalTerminal.promptFor(this, ids, wait, true);
      }
      
      finally
      {
         formatFilter = save;
         
         // here is the extra view (puts the double in double-vision)      
         if (tempTerm != -1)
         {
            LogicalTerminal.switchOutput(tempTerm, false);
            view();
         }
      }
   }
   
   /**
    * Scrolls the lines of the down frame up or down.
    * <p>
    * This method corresponds to these Progress statements:
    * SCROLL
    * <p>
    * The server-side screen buffer will be updated accordingly so it will
    * map the right values for the rows. 
    *
    * @param  fromCurrent
    *         <code>true</code> if FROM-CURRENT option was used
    * @param  upDown
    *         <code>true</code> if scrolling up
    */
   @Override
   public void scroll(boolean fromCurrent, boolean upDown)
   {
      int oldFrameLine = frameLineWorker();
      
      conditionalUpDown();
      LogicalTerminal.scroll(this, fromCurrent, upDown);
      
      // implement the scrolling behavior for the cached buffers
      // no row-drawing is needed because this is implemented on the
      // client-side
      int down  = dynamicDown;
      int start = 0;
      int end   = 0;
      int incr  = 0;

      int frameLine = 1;
      
      if (fromCurrent)
         frameLine = oldFrameLine;
      
      if (upDown)
      {
         incr = 1;
         start = frameLine;
         end = down - 1;
      }
      else
      {
         incr = -1;
         start = frameLine + 1;
         end = down;  
      }
      
      Map<Integer, ScreenBuffer> oldScreenBuffer = screenBufferCache;
      screenBufferCache = new HashMap<Integer, ScreenBuffer>();
      
      for (int i = 1; i < start - 1; i++)
      {         
         ScreenBuffer oldBuf = oldScreenBuffer.get(i);
         addScreenBufferToCache(oldBuf, i);
      }
      
      for (int newkey = start; newkey <= end; newkey++)
      {
         int oldkey = newkey + incr;
         
         ScreenBuffer oldBuf = oldScreenBuffer.get(oldkey);
         addScreenBufferToCache(oldBuf, newkey);
      }
      
      showScreenBuffer(frameLineWorker());
   }

   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    data
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    */
   @Override
   public void set(FrameElement[] data)
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         set(in, data);
      }
      else
      {
         // interactive case
         set(data, null);
      }
   }
   
   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. Data
    * will be copied back from the screen buffer to the user's variables.
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    data
    *           The list of widgets which are to receive input from the
    *           stream.
    */
   @Override
   public void set(Stream in, FrameElement[] data)
   {
      processStreamInput(in, data, true);
   }
   
   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void set(FrameElement[] widgets, EventList wait)
   {
      setWorker(widgets, wait, false);
   }
   
   /**
    * Positions the cursor on a new line in a down frame.
    * This is a conditional call. Multiple conditional up() calls have
    * the effect of a single up(1) call, which is delayed till the nearest
    * terminal I/O operation and can be cancelled by a conditional down()
    * call.
    * <p>
    * This method corresponds to these Progress statements:
    * UP
    */
   @Override
   public void up()
   {
      if (isRedirected())
      {
         // this is a NOP
         return;
      }

      if (pendingUpDown > -1)
      {
         int frameLine =
            LogicalTerminal.getFrameLine(this, pendingUpDown).intValue();
         
         pendingUpDown--;
         
         // non-down frames maintain the pending value but don't swap buffers
         if (!frame.isDown())
            return;
         
         // we add the screen buffer for the previous frame line
         addScreenBufferToCache(this.frameBuf, frameLine);

         // restore the buffer for the next row, but don't send it to the 
         // client
         int nextLine = frameLine - 1;
         frameBuf = getScreenBufferFromCache(nextLine);
      }
   }

   /**
    * Positions the cursor on a new line of the down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * UP n
    *
    * @param  lines
    *         number of iterations of the data to scroll up (positive) or
    *         down (negative).
    */
   @Override
   public void up(double lines)
   {
      up((int) lines);
   }

   /**
    * Positions the cursor on a new line of the down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * UP n
    *
    * @param  lines
    *         number of iterations of the data to scroll up (positive) or
    *         down (negative).
   */
   @Override
   public void up(NumberType lines)
   {
      up(lines.intValue());
   }

   /**
    * This method exists to support a valid form of the Progress 
    * <code>UP</code> statement which has no effect.
    * <p>
    * This method corresponds to these Progress statements:
    * UP STREAM s
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void up(Stream out)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      // otherwise this is a NOP
      return;
   }

   /**
    * If <code>lines</code> is a negative number, the absolute value of that
    * number of new-lines will be written to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * UP STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           If negative, the number of new-lines to output to the stream.
    *           Otherwise, this has no effect.
    */
   @Override
   public void up(Stream out, double lines)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      if (lines > -1)
      {
         // this is a NOP
         return;
      }
      
      up((int) lines, getRemoteStreamId(out));
   }

   /**
    * If <code>lines</code> is a negative number, the absolute value of that
    * number of new-lines will be written to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * UP STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           If negative, the number of new-lines to output to the stream.
    *           Otherwise, this has no effect.
    */
   @Override
   public void up(Stream out, NumberType lines)
   {
      up(out, lines.intValue());
   }

   /**
    * Copies data from the frame elements to each corresponding widget, then
    * enables the list of widgets, shifts into input blocking mode via
    * {@link #waitFor}, after the user's edits are complete the list of 
    * widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real
    * work.  The only difference is that the <code>set</code> is embedded
    * inside a "hidden" <code>TransactionManager</code> block which handles
    * undo/retry on <code>error</code> conditions and undo/leave on 
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    data
    *           The list of data and widgets which is to be updated.
    */
   @Override
   public void update(FrameElement[] data)
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         update(in, data);
      }
      else
      {
         // interactive case
         update(data, null);
      }      
   }
   
   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. The 
    * data passed in will be copied from the user's variables into the screen
    * buffer.  After the input is read from the stream into the screen buffer,
    * that data will be copied back from the screen buffer to the user's
    * variables.  
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.  Unlike the <code>update</code>
    * version in this class which has a "hidden" retry block around the
    * core "set" processing, this streams version has no such feature. Errors
    * will be immediately seen by the caller.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    data
    *           The list of widgets which are to receive input from the
    *           stream.
    */
   @Override
   public void update(Stream in, FrameElement[] data)
   {
      setStaticWidgets(data);
      copyToScreenBuffer(data);
      processStreamInput(in, data, true);
   }

   /**
    * Copies data from the frame elements to each corresponding widget, then
    * enables the list of widgets, shifts into input blocking mode via
    * {@link #waitFor}, after the user's edits are complete the list of 
    * widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real
    * work.  The only difference is that the <code>set</code> is embedded
    * inside a "hidden" <code>TransactionManager</code> block which handles
    * undo/retry on <code>error</code> conditions and undo/leave on 
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    widgets
    *           The list of data and widgets which is to be updated.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void update(FrameElement[] widgets, EventList wait)
   {
      // turning off infinite loop protection
      TransactionManager.disableLoopProtection(null);

      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         update(in, widgets);
         
         return;
      }
      
      // interactive case
      setStaticWidgets(widgets);
      copyToScreenBuffer(widgets);
      
      // the set processing is done inside a "hidden" block that has retry
      // and undo support on errors, no end-key property
      TransactionManager.pushScope("_hiddenUpdateBlock_",
                                   TransactionManager.SUB_TRANSACTION,
                                   false,
                                   false,
                                   false,
                                   false);
      try
      {
         update:
         {
            TransactionManager.blockSetup();
            
            // force the TM to avoid accumulating any contained block's
            // interactions counter (this is a quirk of UPDATE that is not
            // duplicated in SET/PROMPT-FOR); note that only a nested trigger
            // can cause such an accumulation... so we must bypass that case
            // here
            TransactionManager.setRollup(false);
            
            do
            {
               try
               {
                  if (wait == null)
                  {
                     wait = new EventList(true);
                  }
                  wait.setOperation(EventList.UPDATE);
                  
                  // the real work is done here!
                  setWorker(widgets, wait, true);
               }
               
               catch (ErrorConditionException err_except)
               {
                  // special case where we exit if this was caused by the 
                  // END-ERROR key, otherwise we retry
                  if (err_except.isEndError())
                  {
                     throw err_except;
                  }
                  
                  TransactionManager.rollback("_hiddenUpdateBlock_");
                  TransactionManager.triggerRetry(null);
               }
            }
            while (TransactionManager.needsRetry());
            
            if (TransactionManager.isBreakPending())
            {
               break update;
            }
         }
      }
      
      finally
      {
         // we don't need to reset our rollup flag since this hidden block
         // is exiting anyway and that flag only affects accumulation of
         // contained blocks NOT how our current block's interactions counter
         // is accumulated
         
         TransactionManager.popScope();
      }      
   }
   
   /**
    * Makes the frame visible.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW FRAME
    */
   @Override
   public void view()
   {
      conditionalUpDown();

      Stream  out    = UnnamedStreams.output();
      boolean bypass = false;
      
      // do we have an interactive terminal (not redirected output)
      
      if (out == null)
      {
         updateDynamicContent();
         makeFrameVisible();
      }
      else
      {
         bypass = isPageElement(out);
         
         if (bypass)
         {
            LogicalTerminal.registerPageElement(frameId, getPageTop());
            needsPageElementCleanup = true;
         }
      }
      
      if (!bypass)
      {
         LogicalTerminal.view(this, new int[] {frameId});
      }
   }
   
   /**
    * Makes visible specified widgets.
    * 
    * @param   data
    *          Array of FrameElement instances which need to be made visible.
    */
   @Override
   public void view(FrameElement[] data)
   {
      conditionalUpDown();

      if (UnnamedStreams.output() != null || data == null)
      {
         // fallback to old method for stream redirection case
         view();
         return;
      }

      updateDynamicContent();
      makeFrameVisible();

      int sz = 0;
      
      for (int i = 0; i < data.length; i++)
         if (data[i] != null)
            sz++;
      
      int[] ids = new int[sz];
      
      for (int i = 0, j = 0; i < data.length; i++)
      {
         if (data[i] != null) 
            ids[j++] = data[i].getWidget().getId();
      }
      
      LogicalTerminal.view(this, ids);
   }
   
   /**
    * Writes the current screen-buffer contents to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void view(Stream out)
   {
      // No input data means we use the current frame data
      sameViewDataRedirected = true;
      viewWorker(out, null, true);
   }
   
   /**
    * Writes the current screen-buffer contents to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param   data
    *          Array of FrameElement instances which need to be made visible.
    */
   public void view(Stream out, FrameElement[] data)
   {
      // New input data - reset the flag
      sameViewDataRedirected = false;
      viewWorker(out, data, true);
   }
   
   /**
    * Writes the current screen-buffer contents to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           Array of FrameElement instances which need to be made visible.
    * @param    view
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    */
   private void viewWorker(Stream out, FrameElement[] data, boolean view)
   {
      if (out != null)
      {
         // strange case where the error messages are doubled (must be 
         // something deep inside the Progress UI implementation) which occurs
         // for all DISPLAY and any non-frame VIEW that has > 1 widget in the
         // list
         if (!view || (data != null && data.length > 1))
         {
            try
            {
               if (out.protectWrites())
               {
                  // silent error mode is ON (this is NO-ERROR) so the error
                  // output is suppressed and doesn't need to be doubled,
                  // just exit
                  return;
               }
            }
            
            catch (ErrorConditionException ece)
            {
               // ignore because we have to do this again
            }
         }
         
         // normal safety code here, exit in NO-ERROR mode if needed
         if (out.protectWrites())
            return;
      }
      
      conditionalUpDown();

      // if needed, temporarily switch the terminal to a new output
      // destination on the client side
      int sid = getRemoteStreamId(out); 
      
      boolean displayBypass = false;
      boolean register      = false;
      boolean pageTop       = false;

      redirOut = out;
      
      // check if we must register as a page element
      if (isPageElement(out))
      {
         displayBypass = view;
         register      = true;
         pageTop       = getPageTop();
         
         // if we will bypass the view() below, then we need to at least
         // register the page element here
         if (displayBypass)
         {
            LogicalTerminal.registerPageElement(frameId, pageTop, sid);
         }
         
         needsPageElementCleanup = true;
      }
      
      // output the screen-buffer to the current destination
      if (!displayBypass)
      {
         int[] ids = null;
         if (data != null)
         {
            GenericWidget[] wids = widgetListFromFrameElements(data);
            ids = idsFromWidgets(wids);
         }
         else
         {
            ids = new int[] {frameId};
         }

         LogicalTerminal.view(this, ids, sid, register, pageTop, view);
      }
   }
   
   /**
    * Returns the column position of the left corner of the frame within its
    * window.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-COL
    * 
    * @return  current value of the frame column.
    */
   @Override
   public decimal frameCol()
   {
      return LogicalTerminal.getFrameCol(this);
   }

   /**
    * Returns the number of iterations in a down frame.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-DOWN
    * 
    * @return  current value of the frame DOWN value.
    */
   @Override
   public integer frameDown()
   {
      integer fd = LogicalTerminal.getFrameDown(this);
      
      return fd;
   }

   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-LINE
    * 
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is redirected to a stream or if the frame is
    *           not visible, then 0 is returned. If the frame is not a down
    *           frame then 1 will be returned.
    */
   @Override
   public integer frameLine()
   {
      return  new integer(frameLineWorker());
   }

   /**
    * Returns the row position of the upper-left corner of the frame within 
    * its window.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-ROW
    * 
    * @return  current value of the frame row.
    */
   @Override
   public decimal frameRow()
   {
      return LogicalTerminal.getFrameRow(this);
   }

   /**
    * Sets the ATTR-SPACE option.
    * 
    * @param   attr_space
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setAttrSpace(boolean attr_space)
   {
      ((FrameConfig)frame.getConfig()).setAttrSpace(attr_space);
      pushScreenDefinition();
   }
   
   /**
    * Sets the CENTERED option.
    * 
    * @param   centered
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setCentered(boolean centered)
   {
      ((FrameConfig)frame.getConfig()).setCentered(centered);
      pushScreenDefinition();
   }

   /**
    * Sets the display color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setDcolor(Color color)
   {
      ((FrameConfig)frame.getConfig()).setDcolor(color);
      pushScreenDefinition();
   }

   /**
    * Sets the input (prompt-for) color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setPfcolor(Color color)
   {
      ((FrameConfig)frame.getConfig()).setPfcolor(color);
      pushScreenDefinition();
   }

   /**
    * Sets the display color.
    * 
    * @param   cs
    *          color specification
    */
   @Override
   public void setDcolor(ColorSpec cs)
   {
      ((FrameConfig)frame.getConfig()).setDcolor(cs.convert());
      pushScreenDefinition();
   }

   /**
    * Sets the input (prompt-for) color.
    * 
    * @param   cs
    *          color specification
    */
   @Override
   public void setPfcolor(ColorSpec cs)
   {
      ((FrameConfig)frame.getConfig()).setPfcolor(cs.convert());
      pushScreenDefinition();
   }

   /**
    * Sets the frame column as a callback to read the current value
    * calculated by the given expression. This value is honored at the time
    * that the frame is made visible (for the first time or after being
    * hidden).
    * 
    * @param    column
    *           The expression to evaluate for the frame column.
    */
   @Override
   public void setColumn(Resolvable column)
   {
      dynColumn = column;
      
      if (column != null)
      {
         int col = ((NumberType) column.resolve()).intValue() - 1;
         ((FrameConfig)frame.getConfig()).setColumn(col);
      }
   }
   
   /**
    * Sets the frame column as a callback to read the current value stored in
    * the given <code>Accessor</code>. This value is honored at the time that
    * the frame is made visible (for the first time or after being hidden).
    * 
    * @param    column
    *           The instance variable to read for the frame column.
    */
   @Override
   public void setColumn(Accessor column)
   {
      setColumn(new ExpressionWrapper(column));
   }
   
   /**
    * Sets the column position for this frame (statically).
    *
    * @param    column
    *           The 1-based column number at which to start displaying the 
    *           frame.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(NumberType column)
   {
      setColumn(column.intValue());
   }

   /**
    * Sets the column position for this frame (statically).
    *
    * @param    column
    *           The 1-based column number at which to start displaying the 
    *           frame.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(double column)
   {
      if (insideSetup)
      {
         //TODO: add error message for negative values:
         // -1 invalid. Non-negative constant required. (3493)
         if (column > 0 && column < 1)
            column = 1;
         else if (column == 0) // keep default 
            return; 
         else if (column < 0) // ignore it for now
            return; 
      }
      
      int newColumn = ((int) column) - 1;
      FrameConfig config = ((FrameConfig) frame.getConfig());

      config.setPositionChanged(newColumn != config.getColumn());
      config.setColumn(newColumn);

      pushScreenDefinition();

      config.setPositionChanged(false);
   }

   /**
    * Returns the column position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress <code>FRAME:COLUMN</code> 
    * attribute.
    * 
    * @return   The column position for the upper-left corner of the frame.
    */
   @Override
   public integer getColumn()
   {
      int col = LogicalTerminal.getColumn(this);
      return new integer(col);
   }

   /**
    * Returns the row position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress <code>FRAME:ROW</code>
    * attribute.
    * 
    * @return   The row position for the upper-left corner of the frame.
    */
   @Override
   public integer getRow()
   {
      int row = LogicalTerminal.getRow(this);
      return new integer(row);
   }
   
   /**
    * Sets the number of columns.
    * 
    * @param   columns
    *          number of columns
    */
   @Override
   public void setColumns(int columns)
   {
      ((FrameConfig)frame.getConfig()).setColumns(columns);
      pushScreenDefinition();
   }

   /**
    * Get current state of VIEW-AS DIALOG-BOX option.
    *
    * @return  <code>true</code> if config is a DialogBox config.
    */
   public boolean isDialogBox()
   {
      return ((FrameConfig)frame.getConfig()).isDialog();
   }
   
   /**
    * Sets the VIEW-AS DIALOG-BOX option.
    * 
    * @param   dialogBox
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDialogBox(boolean dialogBox)
   {
      ((FrameConfig)frame.getConfig()).setDialog(dialogBox);
      pushScreenDefinition();
   }

   /**
    * Sets the size of a DOWN frame as a callback to read the current value
    * calculated by the given expression.  This value is then honored at the
    * time that the frame is made visible (for the first time or after being
    * hidden).
    *
    * @param    down
    *           The expression to evaluate for the number of data iterations
    *           in the current (DOWN) frame.
    */
   @Override
   public void setDown(Resolvable down)
   {
      dynDown = down;
      frame.setDown(0);
      pushScreenDefinition();
   }
   
   /**
    * Sets the size of a DOWN frame as a callback to read the current value
    * stored in the given <code>Accessor</code>.  This value is then honored
    * at the time that the frame is made visible (for the first time or after
    * being hidden).
    *
    * @param    down
    *           The simple instance variable to read for the number of data
    *           iterations in the current (DOWN) frame.
    */
   @Override
   public void setDown(Accessor down)
   {
      setDown(new ExpressionWrapper(down));
   }

   /**
    * Sets the size of a DOWN frame (statically).
    *
    * @param    down
    *           The number of data iterations in the DOWN frame.
    */
   @Override
   public void setDown(NumberType down)
   {
      setDown(down.intValue());
   }

   /**
    * Sets the size of a DOWN frame (statically).
    *
    * @param    down
    *           The number of data iterations in the DOWN frame.
    */
   @Override
   public void setDown(double down)
   {
      frame.setDown((int) down);
      pushScreenDefinition();
   }

   /**
    * Sets the height of the widget.
    * 
    * @param   height
    *          height of the widget
    */
   @Override
   public void setHeight(int height)
   {
      FrameConfig config = (FrameConfig) frame.getConfig();
      int oldHeight = insideSetup ? height : getHeight().intValue();
      config.setHeightChars(height);
      
      config.setSizeChanged(oldHeight != height);
      pushScreenDefinition();
      config.setSizeChanged(false);
   }

   /**
    * Sets the height of the widget.
    * 
    * @param   height
    *          height of the widget
    */
   @Override
   public void setHeight(integer height)
   {
      setHeight(height.intValue());
   }

   /**
    * Sets the height of the widget.
    *
    * @param   height
    *          height of the widget
    */
   @Override
   public void setHeight(double height)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets the height of the widget.
    *
    * @param   height
    *          height of the widget
    */
   @Override
   public void setHeight(decimal height)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Returns the height of the widget.
    *
    * @return  height of the widget, retrieved from the 
    *          <code>FRAME:HEIGHT-CHARS</code> attribute.
    */
   @Override
   public integer getHeight()
   {
      return new integer(LogicalTerminal.getHeight(this));
   }

   /**
    * Sets the KEEP-TAB-ORDER option.
    * 
    * @param   keepTabOrder
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setKeepTabOrder(boolean keepTabOrder)
   {
      ((FrameConfig)frame.getConfig()).setKeepTabOrder(keepTabOrder);
      pushScreenDefinition();
   }

   /**
    * Sets the THREE-D option.
    *
    * @param threeD
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setThreeD(boolean threeD)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets the NO-BOX option.
    * 
    * @param   noBox
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoBox(boolean noBox)
   {
      ((FrameConfig)frame.getConfig()).setBox(!noBox);
      pushScreenDefinition();
   }

   /**
    * Sets the NO-HELP option.
    * 
    * @param   noHelp
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoHelp(boolean noHelp)
   {
      ((FrameConfig)frame.getConfig()).setNoHelp(noHelp);
      pushScreenDefinition();
   }

   /**
    * Sets the NO-HIDE option.
    * 
    * @param   noHide
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoHide(boolean noHide)
   {
      ((FrameConfig)frame.getConfig()).setNoHide(noHide);
      pushScreenDefinition();
   }

   /**
    * Sets the NO-LABELS option.
    * 
    * @param   noLabels
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoLabels(boolean noLabels)
   {
      ((FrameConfig)frame.getConfig()).setLabels(!noLabels);
      pushScreenDefinition();
   }

   /**
    * Sets the NO-UNDERLINE option.
    * 
    * @param   noUnderline
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoUnderline(boolean noUnderline)
   {
      ((FrameConfig)frame.getConfig()).setNoUnderline(noUnderline);
      pushScreenDefinition();
   }

   /**
    * Sets the NO-VALIDATE option.
    * 
    * @param   noValidate
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoValidate(boolean noValidate)
   {
      ((FrameConfig)frame.getConfig()).setNoValidate(noValidate);
      pushScreenDefinition();
   }

   /**
    * Sets the OVERLAY option.
    * 
    * @param   overlay
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setOverlay(boolean overlay)
   {
      ((FrameConfig)frame.getConfig()).setOverlay(overlay);
      pushScreenDefinition();
   }

   /**
    * Gets the PAGE-BOTTOM option.
    * 
    * @return   <code>true</code> if the option is set to ON.
    */
   public boolean getPageBottom()
   {
      return ((FrameConfig)frame.getConfig()).getPageBottom();
   }

   /**
    * Sets the PAGE-BOTTOM option.
    * 
    * @param   pageBottom
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setPageBottom(boolean pageBottom)
   {
      ((FrameConfig)frame.getConfig()).setPageBottom(pageBottom);
      pushScreenDefinition();
   }

   /**
    * Gets the PAGE-TOP option.
    * 
    * @return   <code>true</code> if the option is set to ON.
    */
   public boolean getPageTop()
   {
      return ((FrameConfig)frame.getConfig()).getPageTop();
   }

   /**
    * Sets the PAGE-TOP option.
    * 
    * @param   pageTop
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setPageTop(boolean pageTop)
   {
      ((FrameConfig)frame.getConfig()).setPageTop(pageTop);
      pushScreenDefinition();
   }

   /**
    * Sets the PARENT option.
    * 
    * @param   widget
    *          New parent widget for frame.
    */
   @Override
   public void setParent(GenericWidget widget)
   {
      ((FrameConfig) frame.getConfig()).setParent(widget.getId());
      pushScreenDefinition();
   }
   
   /**
    * Get the PARENT option.
    * 
    * @return  handle object for the frame handle.
    * 
    */
   @Override
   public handle getParent()
   {
      int id = ((BaseConfig) frame.getConfig()).getParent();
      return LogicalTerminal.getHandleForId(id);
   }
   
   /**
    * Get the first child of the frame which will be the field group that
    * contains all non-header widgets.
    *
    * @return   The first field group.
    */
   @Override
   public GenericWidget getFirstChild()
   {
      // we create this "pseudo-widget" the first time it is needed
      if (fieldGroup == null)
      {
         LinkedList<GenericWidget> list = new LinkedList<GenericWidget>();

         // iterate though all non-header widgets in the proper order (the
         // n2w must return an ordered iterator)
         for (GenericWidget next : n2w.values())
         {
            // SPACE and SKIP aren't real widgets in Progress terms
            if (next instanceof SkipEntity)
               continue;

            // create the specific list of widgets (in the correct order)
            list.add(next);
         }

         fieldGroup = new FieldGroup(list);
      }

      return fieldGroup;
   }
   
   /**
    * Sets the frame retain size.
    * 
    * @param   retain
    *          retain size
    */
   @Override
   public void setRetain(int retain)
   {
      ((FrameConfig) frame.getConfig()).setRetain(retain);
      pushScreenDefinition();
   }
   
   /**
    * Sets the frame row as a callback to read the current value calculated
    * by the given expression. This value is honored at the time that the
    * frame is made visible (for the first time or after being hidden).
    * 
    * @param    row
    *           The expression to evaluate for the frame row.
    */
   @Override
   public void setRow(Resolvable row)
   {
      dynRow = row;
      
      if (row != null)
      {
         int r = ((NumberType) row.resolve()).intValue() - 1;
         ((FrameConfig)frame.getConfig()).setRow(r);
      }
   }
   
   /**
    * Sets the frame row as a callback to read the current value stored in the
    * given <code>Accessor</code>. This value is honored at the time that the
    * frame is made visible (for the first time or after being hidden).
    * 
    * @param    row
    *           The instance variable to read for the frame row.
    */
   @Override
   public void setRow(Accessor row)
   {
      setRow(new ExpressionWrapper(row));
   }
   
   /**
    * Sets the row position for this frame (statically).
    * 
    * @param    row
    *           The 1-based row number at which to start displaying the frame.
    *           This is converted to a 0-based row number for the internal
    *           runtime.
    */
   @Override
   public void setRow(NumberType row)
   {
      setRow(row.intValue());
   }

   /**
    * Sets the row position for this frame (statically).
    * 
    * @param    row
    *           The 1-based row number at which to start displaying the frame.
    *           This is converted to a 0-based row number for the internal
    *           runtime.
    */
   public void setRow(double row)
   {
      if (insideSetup)
      {
         // ROW can be initialized to negative values without errors
         // TODO: investigate deeper meaning of negative values  
         if (row < 1 && row > 0)
            row = 1;
         else if (row == 0) // keep default
            return;
      }
      
      int newRow = ((int) row) - 1;
      FrameConfig config = ((FrameConfig) frame.getConfig());

      config.setPositionChanged(newRow != config.getRow());
      config.setRow(newRow);

      pushScreenDefinition();

      config.setPositionChanged(false);
   }

   /**
    * Sets the frame scroll size.
    * 
    * @param   scroll
    *          scroll size
    */
   @Override
   public void setScroll(int scroll)
   {
      ((FrameConfig)frame.getConfig()).setScroll(scroll);
      pushScreenDefinition();
   }

   /**
    * Sets the SCROLLABLE option.
    * 
    * @param   scrollable
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setScrollable(boolean scrollable)
   {
      ((FrameConfig)frame.getConfig()).setScrollable(scrollable);
      pushScreenDefinition();
   }

   /**
    * Sets the SCROLLABLE option.
    *
    * @param scrollable
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setScrollable(logical scrollable)
   {
      if (scrollable.isUnknown())
         return;
      
      setScrollable(scrollable.booleanValue());
   }
   
   /**
    * Sets the SIDE-LABELS option.
    * 
    * @param   sideLabels
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setSideLabels(boolean sideLabels)
   {
      ((FrameConfig)frame.getConfig()).setSideLabels(sideLabels);
      pushScreenDefinition();
   }

   /**
    * Sets the frame title.
    *
    * @param    title
    *           The title text to use.
    */
   @Override
   public void setTitle(character title)
   {
      setTitle(title.isUnknown() ? "" : title.getValue());
   }

   /**
    * Sets the frame title.
    *
    * @param    title
    *           The title text to use.
    */
   @Override
   public void setTitle(String title)
   {
      FrameConfig fc = (FrameConfig)frame.getConfig();
      fc.setTitle(title);
      pushScreenDefinition();
   }

   /**
    * Sets the frame title and title color.
    *
    * @param    title
    *           The title text to use.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setTitle(character title, ColorSpec titleColor)
   {
      setTitle(title.isUnknown() ? "" : title.getValue(), titleColor);
   }

   /**
    * Sets the frame title and title color.
    *
    * @param    title
    *           The title text to use.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setTitle(String title, ColorSpec titleColor)
   {
      FrameConfig fc = (FrameConfig)frame.getConfig();
      fc.setTitle(title);
      fc.setTitleDcolor(titleColor);
      pushScreenDefinition();
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback to the given
    * parameter).  The value at the instant of this call is not the value that 
    * will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The simple variable or field reference that contains the
    *           title text to be used.
    */
   @Override
   public void setDynamicTitle(BaseDataType title)
   {
      setDynamicTitle(new AccessorWrapper(title));
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    dyntitle
    *           The simple variable or field reference that contains the
    *           title text to be used.
    */
   @Override
   public void setDynamicTitle(Accessor dyntitle)
   {
      class TitleExpr
      extends CharacterExpression
      {
         private Accessor title = null;
         
         /**
          * Construct an instance for given {@link Accessor} instance.
          * 
          * @param    title
          *           Dynamic title variable.
          */
         public TitleExpr(Accessor title)
         {
            this.title = title;
         }

         @Override
         public character execute()
         {
            return (character) title.get();
         }
      }
      
      setDynamicTitle(new TitleExpr(dyntitle));
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The expression callback that generates the title text to be
    *           used.
    */
   @Override
   public void setDynamicTitle(CharacterExpression title)
   {
      dynTitle = title;
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback to the given
    * parameter).  The value at the instant of this call is not the value that 
    * will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The simple variable or field reference that contains the
    *           title text to be used.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setDynamicTitle(BaseDataType title, ColorSpec titleColor)
   {
      setDynamicTitle(new AccessorWrapper(title), titleColor);
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The simple variable or field reference that contains the
    *           title text to be used.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setDynamicTitle(Accessor title, ColorSpec titleColor)
   {
      setDynamicTitle(title);
      setTitleColor(titleColor);
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The expression callback that generates the title text to be
    *           used.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setDynamicTitle(CharacterExpression title,
                               ColorSpec           titleColor)
   {
      setDynamicTitle(title);
      setTitleColor(titleColor);
   }

   /**
    * Sets the color for the frame title.
    *
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setTitleColor(ColorSpec titleColor)
   {
      ((FrameConfig)frame.getConfig()).setTitleDcolor(titleColor);
      pushScreenDefinition();
   }

   /**
    * Sets the TOP-ONLY option.
    *
    * @param topOnly
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setTopOnly(boolean topOnly)
   {
      ((FrameConfig)frame.getConfig()).setTopOnly(topOnly);
      pushScreenDefinition();
   }

   /**
    * Sets the width of the widget.
    *
    * @param width
    *        width of the widget
    */
   @Override
   public void setWidth(int width)
   {
      FrameConfig config = (FrameConfig) frame.getConfig();
      int oldWidth = insideSetup ? width : getWidth().intValue();
      config.setWidthChars(width);
      
      config.setSizeChanged(oldWidth != width);
      pushScreenDefinition();
      config.setSizeChanged(false);
   }

   /**
    * Sets the width of the widget.
    *
    * @param width
    *        width of the widget
    */
   @Override
   public void setWidth(integer width)
   {
      setWidth(width.intValue());
   }

   /**
    * Sets the width of the widget.
    *
    * @param width
    *        width of the widget
    */
   @Override
   public void setWidth(double width)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets the width of the widget.
    *
    * @param width
    *        width of the widget
    */
   @Override
   public void setWidth(decimal width)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Returns the width of the widget.
    *
    * @return  width of the widget, retrieved from the 
    *          <code>FRAME:WIDTH-CHARS</code> attribute.
    */
   @Override
   public integer getWidth()
   {
      return new integer(LogicalTerminal.getWidth(this));
   }

   /**
    * Returns the width of the widget in pixels. Corresponds Progress
    * WIDTH-PIXELS attribute.
    *
    * @return The width of the widget in pixels.
    */
   @Override
   public integer getWidthPixels()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setWidthPixels(integer widthPixels)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Returns the horisontal position of the frame. Corresponds Progress X
    * attribute.
    *
    * @return the horisontal position of the frame.
    */
   @Override
   public integer getX()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Returns the vertical position of the frame. Corresponds Progress Y
    * attribute.
    *
    * @return the vertical position of the frame.
    */
   @Override
   public integer getY()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets the horizontal position of the frame (Progress X
    * attribute).
    *
    * @param x
    *        horizontal position to be set.
    */
   @Override
   public void setX(NumberType x)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets the vertical position of the frame (Progress Y
    * attribute).
    *
    * @param y
    *        vertical position to be set.
    */
   @Override
   public void setY(NumberType y)
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets the default button for this frame.
    *
    * @param defaultButton
    *        button widget that works as the default
    */
   @Override
   public void setDefaultButton(ButtonWidget defaultButton)
   {
      int id = getWidgetId(defaultButton);
      
      if (id == -1)
         return;

      ((FrameConfig) frame.getConfig()).setDefaultButton(id);
      pushScreenDefinition();
   }

   /**
    * Sets the widget which will be used to display the description for 
    * any given statistic that is being automatically aggregated by the
    * {@link #display} processing.
    *
    * @param    widget
    *           The widget to use as the statistics column.
    */
   @Override
   public void setStatisticsColumn(GenericWidget widget)
   {
      statsColumn = widget;
   }
   
   /**
    * Notifies the frame about entering a new frame scope.  This may occur
    * more than once for a given frame due to retries or (more likely) due
    * to a frame being scoped to a block that is nested inside a loop. In
    * the case where this is a shared frame, this will be called more than
    * once while a previous frame scope is still open.
    * <p>
    * This method registers a callback for the block exit event. The callback
    * finalizes the frame and all contained widgets.
    * <p>
    * On block iterate or retry, frame state will also be maintained.
    */
   @Override
   public void openScope()
   {
      // push down any pending screen definitions
      LogicalTerminal.processDeferredPush();
      
      // register for iterate, retry and finished notifications
      TransactionManager.registerFinalizable(this, false);
      
      // associate this frame with a scope in the LT (this is safe to do
      // multiple times and must be safe because of multiple nested calls
      // to this method)
      
      if (scopeNesting == 0)
      {
         LogicalTerminal.registerFrameInScope(this);
         
         // reset frame to its initial state
         resetFrameBuf();
         pendingUpDown = 0;
         hadView       = false;
         canHide       = true;
         hadCondition  = false;

         // the screen buffer cache needs also to be cleared
         screenBufferCache.clear();
      }
      
      LogicalTerminal.signalFrameScope(this, MarkEntry.OPEN_SCOPE);

      scopeNesting ++;
      
      // mark this frame as a valid frame
      scopeValid = true;
   }

   /**
    * Cleanup that must occur at the end of the block to which the frame is
    * scoped.
    */
   @Override
   public void finished()
   {
      // If we previously bypassed the clean up - do this now
      if (hadView && !frame.isDown() &&
          isOutRedirected() && sameViewDataRedirected)
         frameBuf.reset();

      if (hadView && isRedirected() || (isOutRedirected() && !isOutTerm()))
      {
         if (!isFlushed() && getRedirOut() != null)
            getRedirOut().flush();
         flushed = true;
      }
      
      // aggregation needs this to finish aggregation processing 
      if (aggregateHandler != null)
      {
         aggregateHandler.finished();
      }
      
      // clear the list, in case no iteration was performed
      deferredAggregators = null;

      scopeNesting--;
      
      LogicalTerminal.signalFrameScope(this, MarkEntry.CLOSE_SCOPE);
      
      // WARNING: can there be issues with shared frames that have multiple
      //          openScope() calls?  In that case it is possible we may
      //          clear something registered in an outer scope.  WATCH THIS!
      
      if (headers != null)
      {
         headers = null;
         frameBuf.clearHeaders();
      }
      
      if (needsPageElementCleanup)
      {
         LogicalTerminal.forceDeregisterPageElement(frameId, getPageTop());
      }
   }
   
   /**
    * Final cleanups when frame completely leaves outer execute() scope (only
    * occurs once for a frame, even if this is a shared frame which has
    * nested <code>openScope()</code> invocations). 
    */
   public void destroy()
   {
      // completely remove the frame, even if it is a shared frame since this
      // is the last use of it and it is really going out of scope now
      scopeValid = false;

      // Destroy frame at client. Deregistration will happen later,
      // when client will actually hide frame and deregister it. 
      LogicalTerminal.destroyFrame(this);
      
      // reset redirection state
      redirOut = null;
   }
   
   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to iterate and attempt another pass.  This
    * provides an opportunity to preserve state from the previous pass and
    * make preparations for the coming pass.
    */
   @Override
   public void iterate()
   {
      if (deferredAggregators != null)
      {
         processAggregates(deferredAggregators, false);
         deferredAggregators = null;
      }

      if (aggregateHandler != null)
      {
         aggregateHandler.iterate();
      }

      canHide = true;
      resetFrameBuf();
      
      if (isOutRedirected())
         down(getRedirOut());
      else
         down();
   }

   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to retry and re-attempt the current iteration.
    * This provides an opportunity to clear or reset state from the previous
    * pass.
    */
   @Override
   public void retry()
   {
      if (!hadCondition )
      {
         resetFrameBuf();
         LogicalTerminal.resetFrame(this);
      }
      hadCondition = false;
   }
   
   /**
    * Marks the beginning of the editing block.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This
    *           is used to ensure that the proper semantics of update or set
    *           are honored at editing block start (optionally copying data
    *           into the screen buffer, enabling widgets and view).
    * @param    type
    *           Specifies if this editing block must operates like
    *           {@link #promptFor} (no copying to or from the screen buffer),
    *           {@link #set} (no copying to the screen buffer) or 
    *           {@link #update} (copying both directions).  Use the constants
    *           PROMPT_FOR, SET and UPDATE to specify this parameter. 
    */
   @Override
   public void startEditingMode(FrameElement[] list, int type)
   {
      // create a default GO-ON list
      EventList eventList = new EventList(true);

      startEditingMode(list, type, eventList);
   }
   
   /**
    * Marks the beginning of the editing block with caller control over the
    * exit conditions for the loop.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This
    *           is used to ensure that the proper semantics of update or set
    *           are honored at editing block start (optionally copying data
    *           into the screen buffer, enabling widgets and view).
    * @param    type
    *           Specifies if this editing block must operates like
    *           {@link #promptFor} (no copying to or from the screen buffer),
    *           {@link #set} (no copying to the screen buffer) or 
    *           {@link #update} (copying both directions).  Use the constants
    *           PROMPT_FOR, SET and UPDATE to specify this parameter. 
    * @param    wait
    *           List of events to wait for.
    *
    */
   @Override
   @SuppressWarnings("unused")
   public void startEditingMode(FrameElement[] list, 
                                int            type,
                                EventList      wait)
   {
      // this object will register itself for exit notification from the 
      // current block, if the exit is abnormal it will trigger cleanup
      EditingCleaner ec = new EditingCleaner(list);
      
      synchronized (editLock)
      {
         if (currentOp != NONE)
         {
            throw new IllegalStateException("Cannot nest editing blocks!");
         }
         
         currentOp = type;
      }
      
      // process text grouping (word wrap mode)
      textGroup  = textGroup(list);
      
      if (type == UPDATE)
      {
         copyToScreenBuffer(list);
      }
      
      GenericWidget[] wids = widgetListFromFrameElements(list);
      
      int[] ids = validIds(wids);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      conditionalUpDown();      
      
      LogicalTerminal.promptFor(this, ids, wait, false);
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    list 
    *           List of widgets.
    */
   @Override
   public void underline(GenericWidget[] list)
   {
      int[] ids = validIds(list);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      underline(ids);
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    list 
    *           List of widget ids.
    */
   @Override
   public void underline(int[] list)
   {
      if (!afterDown)
         down();
      
      conditionalUpDown();
      
      // drive the process
      LogicalTerminal.underline(this, list);
      down();
      afterUnderline = true;
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    stream
    *           An explicit stream which will be used for the output.
    * @param    list 
    *           List of widgets.
    */
   @Override
   public void underline(Stream stream, GenericWidget[] list)
   {
      int[] ids = validIds(list);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      underline(stream, ids);
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    stream
    *           An explicit stream which will be used for the output.
    * @param    list 
    *           List of widget ids.
    */
   @Override
   public void underline(Stream stream, int[] list)
   {
      if (stream != null && stream.protectWrites())
      {
         return;
      }
      
      // temporarily switch the terminal to a new output destination
      int id      = getRemoteStreamId(stream); 
      int tempOut = LogicalTerminal.switchOutput(id, true, true);
      
      try
      {
         // because of the internal down/conditionalUpDown behavior of this
         // next method call, we cannot easily move the switchOutput() call
         // to the client side
         underline(list);
      }
      
      finally
      {
         // put the old terminal back (either interactive or a different
         // redirection)
         LogicalTerminal.switchOutput(tempOut, false);
      }
   }
   
   /**
    * Waits for a default list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR </b>
    */
   @Override
   public void waitFor()
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR </b>
    * 
    * @param   el
    *          List of events to wait for.
    */
   @Override
   public void waitFor(EventList el)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget </b>
    * 
    * @param   el
    *          List of events to wait for.
    * @param   focusWidget
    *          Widget which is in focus.
    */
   @Override
   public void waitFor(EventList el, GenericWidget focusWidget)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el, focusWidget);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR PAUSE n </b>
    * 
    * @param   el
    *          List of events to wait for.
    * @param   seconds
    *          Time to wait for event.
    */
   @Override
   public void waitFor(EventList el, int seconds)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el, seconds);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    * 
    * @param   el
    *          List of events to wait for.
    * @param   focusWidget
    *          Widget which is in focus.
    * @param   seconds
    *          Time to wait for event.
    */
   @Override
   public void waitFor(EventList el, GenericWidget focusWidget,
                       int seconds)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el, focusWidget, seconds);
   }
      
   /**
    * Blocks until a next UI event is available (returns immediately if an
    * event is already available).
    */
   @Override
   public void waitForNextKey()
   {
      LogicalTerminal.waitForNextKey();
   }
   
   /**
    * Marks the end of the editing block when triggered by any abnormal exit
    * from the block on the server.  This method does NOT need to be called
    * if the exit is naturally triggered via {@link #continueEditing}. This
    * method allows the client's state to be cleaned up in the abnormal
    * termination case.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This
    *           is used to ensure that the proper semantics of update or set
    *           are honored at editing block exit (disabling widgets and
    *           assigning data back).
    * @param    normal
    *           <code>true</code> if this was triggered as the result of
    *           the natural pending GO exit from the editing block (if this 
    *           was the result of a call to {@link #continueEditing}).
    *           <code>false</code> if this was due to an abnormal exit and
    *           the client's state needs cleanup.
    * @param    cleanup
    *           <code>true</code> if the client side cleanup is needed.
    */
   public void stopEditingMode(FrameElement[] list,
                               boolean        normal,
                               boolean        cleanup)
   {
      GenericWidget[] wids = widgetListFromFrameElements(list);
      
      int[] ids = validIds(wids);
      
      // only notify the client of the stop and disable if we actually
      // enabled and started editing mode in the first place, if there are
      // no widgets specified, then we never entered editing mode in the
      // first place
      if (ids != null)
      {
         if (cleanup)
         {
            // clear the client's state
            LogicalTerminal.stopEditingMode();
         }
         
         LogicalTerminal.disable(this, ids);
      }
      
      textUngroup(textGroup);
      textGroup = null;
      
      synchronized (editLock)
      {
         if (ids != null)
         {
            // only copy changes back if we exited normally (and we are one
            // of the op types that copy back)
            if (normal && (currentOp == SET || currentOp == UPDATE))
            {
               copyFromScreenBuffer(list, (currentOp == UPDATE));
            }            
         }         
         
         currentOp = NONE;
      }
   }
   
   /**
    * Checks to see whether a pending GO event is about to end the editing
    * block.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This
    *           is used to ensure that the proper semantics of update or set
    *           are honored at editing block exit (disabling widgets and
    *           assigning data back).
    *
    * @return   <code>true</code> if there is NO pending GO.
    */
   @Override
   public boolean continueEditing(FrameElement[] list)
   {
      boolean cont = LogicalTerminal.continueEditing();
      
      if (!cont)
      {
         stopEditingMode(list, true, false);
      }
      
      return cont;
   }

   /**
    * Executes the validation tests for every supported field-level widget
    * in the frame (except the browse, which you must validate explicitly).
    * Corresponds Progress VALIDATE method.
    *
    * @return <code>true</code> if validation was successfull.
    */
   @Override
   public logical validate()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Sets custom colors for output fields, input fields, or both in
    * all specified widgets.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>COLOR DISPLAY [PROMPT]</b> <br>
    * <b>COLOR PROMPT</b>
    * 
    * @param    display
    *           A color specification for regular output or <code>null</code>
    *           if no change is required.
    * @param    prompt
    *           A color specification for input or <code>null</code> if
    *           no change is required.
    * @param    widgets
    *           Array of target widgets.  All elements must be in the same
    *           frame.
    */
   @Override
   public void setColors(ColorSpec       display,
                         ColorSpec       prompt, 
                         GenericWidget[] widgets)
   {
      // honor any pending conditional down/up processing
      conditionalUpDown();
      
      // batch up the changes (defer processing until later)
      batch(true);
      
      for (GenericWidget widget : widgets)
      {
         widget.saveColors();
         
         if (display != null)
            widget.colorDisplay(display);
         
         if (prompt != null)
            widget.colorPrompt(prompt);
      }
      
      // suppress redrawing of the entire frame
      frameDef.setSuppressRedraw(true);
      
      // push the changes
      batch(false, true);
      
      // restore colors in DOWN frames
      if (frame.isDown())
      {
         for (GenericWidget widget : widgets)
         {
            if (!(widget instanceof FrameWidget))
               widget.restoreColors();
         }
      }
   }

   /**
    * Sets custom colors for the given output field, input field, or both.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>COLOR DISPLAY [PROMPT]</b> <br>
    * <b>COLOR PROMPT</b>
    * 
    * @param    display
    *           A color specification for regular output or <code>null</code>
    *           if no change is required.
    * @param    prompt
    *           A color specification for input or <code>null</code> if
    *           no change is required.
    * @param    widget
    *           The target widget.
    */
   @Override
   public void setColors(ColorSpec     display,
                         ColorSpec     prompt, 
                         GenericWidget widget)
   {
      setColors(display, prompt, new GenericWidget[] {widget});
   }

   /**
    * Reports if this frame object is valid for use.
    * <p>
    * The validity of the frame object is determined solely by being in 
    * scope.
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean isValid()
   {
      return scopeValid;
   }

   /**
    * Reports if this object is unknown.<p>
    * This implementation always returns <code>false</code>.
    *
    * @return   <code>false</code>.
    */
   @Override
   public boolean isUnknown()
   {
      return false;
   }

   /**
    * Gets the unique numeric ID of this frame.
    *
    * @return  frame ID
    */
   @Override
   public int getFrameId()
   {
      return frameId;
   }

   /**
    * Gets the screen buffer reference of this frame.
    *
    * @return  screen buffer
    */
   public ScreenBuffer getFrameBufferRef()
   {
      updateDynamicContent();
      return frameBuf;
   }

   /**
    * Gets the differential screen buffer of this frame.
    *
    * @return  screen buffer
    */
   public ScreenBuffer getFrameBuffer()
   {
      updateDynamicContent();
      return frameBuf.getChanged();
   }

   /**
    * Gets the frame widget of this frame.
    * 
    * @return  <code>FrameWidget</code>
    */
   public FrameWidget getFrameWidget()
   {
      return frame;
   }

   /**
    * Request indexes of selected items in RadioSet, ComboBox or SelectionList
    * widgets.
    * 
    * @param   widget
    *          Widget for which request should be performed.
    * 
    * @return  array of 0-based indexes of selected items or <code>null</code>
    *          if specified widget does not exists or does not support such a
    *          request.
    */
   public int[] getSelection(GenericWidget widget)
   {
      return LogicalTerminal.getSelection(widget.getId());
   }

   /**
    * Obtains the type of widget description.
    *
    * @return   Always "FRAME".
    */
   @Override
   public character getType()
   {
      return new character("FRAME");
   }

   /**
    * Get widget name for the given widget.
    * 
    * @param   widget
    *          Widget to look name for.
    * 
    * @return  Widget name.
    */
   public String getName(GenericWidget widget)
   {
      if (widget.equals(frame))
      {
         return new String(frameName);
      }

      for (Iterator<String> iter = n2w.keySet().iterator(); iter.hasNext();)
      {
         String name = iter.next();
         
         if (n2w.get(name) == widget)
            return name;
      }
      
      return "Unknown";
   }

   /**
    * Get original widget name for the given widget.
    * 
    * @param   widget
    *          Widget to look name for.
    * 
    * @return  original widget name.
    */
   public String getOriginalName(GenericWidget widget)
   {
      return n2n.get(getName(widget));
   }

   /**
    * Gets a set of widgets of this frame that require validation.
    * 
    * @return  a <code>Set</code> of widgets, which may be empty. 
    */
   public Set<GenericWidget> getValidationSet()
   {
      Set<GenericWidget> set = new HashSet<GenericWidget>();
      Iterator<String> iter = n2w.keySet().iterator();
      
      while (iter.hasNext())
      {
         GenericWidget gw = n2w.get(iter.next());
         
         if (gw.isValidatable())
            set.add(gw);
      }
      
      return set;
   }

   /**
    * Mark frame as viewed.
    */
   public void setHadView()
   {
      hadView = true;
      canHide = false;
      flushed = false;
   }
   
   /**
    * Get state of internal flag which is set to <code>true</code> if frame
    * was viewed at client at least once.
    * 
    * @return  <code>true</code> if frame was viewed.
    */
   public boolean isViewed()
   {
      return hadView;
   }
   
   /**
    * Check if frame can hide other frames.
    * 
    * @return <code>true</code> if frame can hide other frames.
    */
   public boolean isCanHide()
   {
      return canHide;
   }
   
   /**
    * This method implements MOVE-AFTER-TAB function.
    * 
    * @param   widget
    *          Widget reference.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public boolean moveAfterTab(GenericWidget widget)
   {
      throw new UnsupportedOperationException("Not implemented");
   }
   
   /**
    * Mark beginning of trigger processing. 
    */
   void triggerEnter()
   {
      triggerActive++;
   }

   /**
    * Mark end of trigger processing. 
    */
   void triggerLeave()
   {
      triggerActive--;
      silentFormatErrors = false;
   }
   
   /**
    * Conditionally pushes the screen definition instance down to the client. 
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of 
    * <code>LogicalTerminal</code> is called to send it down to the client.
    */
   void pushScreenDefinition()
   {
      pushScreenDefinition(true);
   }
   
   /**
    * Conditionally pushes the screen definition instance down to the client. 
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of 
    * <code>LogicalTerminal</code> is called to send it down to the client.
    * 
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push.
    */
   void pushScreenDefinition(boolean force)
   {
      // batching is in progress, do not push individual changes
      if (batch)
         return;

      // time to push the changes down
      if (frameDef == null)
      {
         // create a new frameDef
         frameDef = new ScreenDefinition(frameId);
         
         // add the frame config as the first item
         FrameConfig fc = (FrameConfig)frame.getConfig();
         
         frameDef.addConfig(fc, frameId);

         // add headers first
         
         ComponentConfig cfg = null;
         
         Iterator<String> iter = n2h.keySet().iterator();
         
         while (iter.hasNext())
         {
            GenericWidget gw = n2h.get(iter.next());
            
            cfg = gw.getConfig();
            cfg.setFrameId(frameId);
            
            frameDef.addConfig(cfg);
         }
         
         // add all the widgets in the order of their definition
         
         for (int i = 0; i < widgets.length; i++)
         {
            cfg = widgets[i].getConfig();
            cfg.setFrameId(frameId);
            frameDef.addConfig(cfg);
         }
         
         if (isShared() && frame.isVisible().booleanValue())
            frameDef.setSuppressRedraw(true);
      }
      
      // push down the screen definition
      if (force)
      {
         // synchronous push
         ScreenDefinition[] sd = new ScreenDefinition[] {frameDef};
         LogicalTerminal.pushScreenDefinition(sd);
      }
      else
      {
         // deferred push
         LogicalTerminal.deferPush(frameDef);
      }
   }
   
   /**
    * Sets a new ScreenBuffer for this frame.
    * 
    * @param   sb
    *          new ScreenBuffer for this frame
    * @param   allow
    *          <code>true</code> to allow the switching of screen buffers if
    *          the frame is a non-redirected down frame and the frame-line is
    *          detected as changing.  Otherwise this switching will be
    *          disabled.
    */
   void setScreenBuffer(ScreenBuffer sb, boolean allow)
   {
      if (allow)
      {
         int oldline = frameBuf.getLine();
         int newline = sb.getLine();
         
         // check if the frame line is different than before AND this is a
         // down frame that is not redirected
         if (oldline != newline && frame.isDown() && !isRedirected())
         {
            // the buffer cache is 1-based so we can only dereference an
            // existing screen buffer if the index is 1 or above
            if (newline > 0)
            {
               // switch screen buffers because the client switched frame
               // lines
               frameBuf = getScreenBufferFromCache(newline);
            }
         }
      }
      
      frameBuf.mergeChanged(sb);
      frameBuf.resetChanged();
      KeyReader.setLastKey(sb.getKeyCode());
   }
   
   /**
    * Returns -1 if the widget name is invalid or <code>null</code>, otherwise 
    * it returns the assigned widget ID.
    *
    * @param    widgetName
    *           A widget name.
    *
    * @return   The numeric ID assigned to the widget or -1 if no such widget
    *           exists.
    */
   int getWidgetId(String widgetName)
   {
       return getWidgetId(n2w.get(widgetName));
   }
   
   /**
    * Returns -1 if the widget is invalid or <code>null</code>, otherwise 
    * it returns the assigned widget ID.
    *
    * @param    widget
    *           A widget instance in the frame.
    *
    * @return   The numeric ID assigned to the widget or -1 if no such widget
    *           exists.
    */
   int getWidgetId(GenericWidget widget)
   {
      if (widget != null)
      {
         return widget.getId();
      }
      
      return -1;
   }

   /**
    * Looks up the widget by its ID.
    * 
    * @param    id
    *           The numeric ID assigned to the widget.
    * 
    * @return   The associated widget instance or <code>null</code> if a
    *           widget by that ID does not exist in this frame.
    */
   GenericWidget getWidgetForId(int id)
   {
      int idx = (id - (frameId + 1));
      
      if (idx >= 0 && idx < widgets.length)
      {
         return widgets[idx];
      }
      
      return null;
   }

   /**
    * Looks up the header widget by its ID.
    * 
    * @param    id
    *           The numeric ID assigned to the header widget.
    * 
    * @return   The associated header widget instance or <code>null</code> if
    *           a widget by that ID does not exist in this frame.
    */
   GenericWidget getHeaderWidgetForId(int id)
   {
      Iterator<GenericWidget> itr = n2h.values().iterator();
      while (itr.hasNext())
      {
         GenericWidget w = itr.next();
         if (w.getId() == id)
         {
            return w;
         }
      }

      return null;
   }
   
   /**
    * Let frame know that there was a condition during frame processing. 
    */
   void setCondition()
   {
      hadCondition = true;
   }
   
   /**
    * Read the next field from the given stream using any whitespace as the 
    * delimiter and validate the data against the widget's format string 
    * before assigning the data into the screen-buffer and back to the
    * user's variable.  Screen-buffer and widget processing can be bypassed
    * by passing <code>null</code> for these values and in this case, the
    * stream reading will not operate upon any of the user interface state.
    * <p>
    * If the current line buffer is empty, a new line will be read from the
    * stream.
    * <p>
    * In the following conditions, no assignment will be made:
    * <ul>
    *   <li> a field is requested after the entire line has already been
    *        processed
    *   <li> an unquoted hyphen is used as the contents of the field AND the 
    *        delimiter is NOT a hyphen (the index is still incremented past
    *        the field)
    *   <li> the first char of a field starts with a double quote but there
    *        is no closing quote (the index is incremented to the end of the
    *        buffer)
    * </ul>
    * <p>
    * After all field reads for a given line are complete, 
    * <code>Stream.resetCurrentLine</code> must be called to enable input for
    * the next line.  Otherwise, all subsequent reads will continue trying to
    * read from the current line. This will result in reading data left on
    * the current line or reading nothing if the line has already been fully 
    * read.  <code>Stream.resetCurrentLine</code> notifies the stream
    * environment that it should move to processing the next line.
    *
    * @param    in
    *           The input stream from which to read.
    * @param    data
    *           The variable to process.  If <code>null</code>, then the
    *           field must be read but the data should be discarded.
    * @param    frameBuf
    *           The frame buffer to update.  Should be <code>null</code> if
    *           there is no frame buffer to update.
    * @param    wid
    *           The widget to process.  Should be <code>null</code> if the
    *           <code>data</code> parameter is <code>null</code> or if there
    *           is no associated widget.
    * @param    fmt
    *           Format string to use for validation.  If <code>null</code> 
    *           obtain it from the widget (if specified) or use the default
    *           for the data type if there is no widget.
    * @param    assign
    *           <code>true</code> if the read data should be assigned back to
    *           the caller's <code>data</code> instance.
    *
    * @return   <code>true</code> if no error occurred, <code>false</code> if
    *           further processing should be aborted due to a failure in this
    *           method.
    *           
    * @throws   EndConditionException 
    * @throws   ErrorConditionException 
    */
   @SuppressWarnings("null")
   static boolean getField(Stream        in,
                           Accessor      data,
                           ScreenBuffer  frameBuf,
                           GenericWidget wid,
                           String        fmt,
                           boolean       assign)
   throws EndConditionException,
          ErrorConditionException
   {   
      BaseDataType var = (data == null ? null : data.get().duplicate());
      String       result = in.readFieldWorker(var);

      if (result != null)
      {         
         // first try to get the format string
         if (fmt == null && wid != null)
         {
            fmt = wid.getFormat();
         }
         
         // fallback plan
         if (fmt == null)
         {
            fmt = var.defaultFormatString();
         }
         
         int len = fmt.length();
         
         // format string prep
         if (var instanceof character)
         {
            // character format strings need expansion
            fmt = character.preprocessFormatString(fmt);
            len = fmt.length();
         }
         else if (var instanceof logical)
         {
            // logical format strings
            len = logical.maximumStringSize(fmt);
         }
         
         // truncate the text if the read data can't fit into the format
         if (result.length() > len)
         {
            result = result.substring(0, len);
            
            if (var instanceof logical)
            {
               // no error in this case keep going
            }
            else
            {
               // for character, we can save the data into the screen buffer
               // like Progress does (we can't do this for the other data
               // types at this time)
               if (var instanceof character &&
                   frameBuf != null         &&
                   wid      != null)
               {
                  frameBuf.putWidgetValue(wid.getId(),
                                          new character(result));
               } 
               
               // throw the error
               String errmsg = "File input is too long: " + result; 
               ErrorManager.recordOrThrowError(557, errmsg);
               
               // if we are in silent error mode return
               return false;
            }
         }
         
         // handle unknown value in a generic way
         if (result.equals("?"))
         {
            var.setUnknown();
         }
         else
         {         
            // convert the value to the proper type
            if (var instanceof integer)
            {
               ((integer) var).assign(new integer(result));
            }
            else if (var instanceof decimal)
            {
               ((decimal) var).assign(new decimal(result));
            }
            else if (var instanceof date)
            {
               ((date) var).assign(new date(result));
            }
            else if (var instanceof logical)
            {
               ((logical) var).assign(new logical(result, fmt));
            }
            else if (var instanceof character)
            {
               ((character) var).assign(new character(result));
            }            
         }
         
         // assign the value to the screen buffer
         if (frameBuf != null && wid != null)
         {
            frameBuf.putWidgetValue(wid.getId(), var);
         }
         
         // assign back to the caller's variable
         if (assign)
         {
            data.set(var);
         }
      }      
      
      return true;
   }
   
   /**
    * Reset frame buffer. Make sure that static widgets are not reset if frame
    * was not displayed yet.
    * <p>
    * NOTE: Current approach assumes that frame is either, at least once
    * displayed or not displayed but only static widgets are assigned. This
    * may result to differences in behavior with Progress when regular
    * (non-static) widget is updated, frame is reset and only then is
    * displayed.
    */
   private void resetFrameBuf()
   {
      // Frame buffer resetting is not required if we use the current data
      // from one iteration to another
      if (hadView && !sameViewDataRedirected &&
          (frame.isDown() || !isOutRedirected()))
         frameBuf.reset();
   }

   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    def
    *           <code>true</code> means that default values will be used
    *           if screen buffer does not contain updated values.
    */
   private void setWorker(FrameElement[] widgets, EventList wait, boolean def)
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         set(in, widgets);
      }
      else
      {
         // interactive case
         if (wait == null)
         {
            wait = new EventList(true);
         }
         wait.setOperation(EventList.SET);
         
         boolean retry;
         int i = 0;
         
         do
         {
            promptFor(widgets, wait);

            retry = false;
            
            if (finalCheck)
            {
               finalCheck = false;
               
               for (i = 0; i < widgets.length; i++)
               {
                  if (widgets[i] == null)
                     continue;
                  
                  GenericWidget widget = widgets[i].getWidget();
                  String format = widget.getFormat();
                  Object value = frameBuf.getWidgetValue(widget.getId());
                  
                  if (value == null || !frameBuf.isEntered(widget.getId()))
                     continue;
                  
                  // finally perform check
                  ErrorEntry res = DisplayFormatFactory.checkFormat((BaseDataType) value, 
                                                                    format);

                  if (res == null)
                     continue;
                  
                  LogicalTerminal.message(res, 
                                          new ErrorEntry(143, 
                                                         configClass.getSimpleName() + 
                                                         ": Unable to evaluate " +
                                                         "field for assignment"));
                  retry = true;
                  break;
               }
            }
            
            if (retry)
            {
               // i points to failed widget
               Utils.scroll(widgets, i);
            }
         }
         while (retry);
         
         // get the value in the screen buffer for each widget using the widget
         // id as the index, then assign this back to the variable from the 
         // caller
         copyFromScreenBuffer(widgets, def);
      }
   }
   
   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-LINE
    * 
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is redirected to a stream or if the frame is
    *           not visible, then 0 is returned. If the frame is not a down
    *           frame then 1 will be returned.
    */
   private int frameLineWorker()
   {
      int line = -1;
      
      /*
      if (frameBuf != null)
      {
         line = frameBuf.getLine();
      }
      */
      
      // taking a trip to the client for this information is REALLY
      // expensive so we should avoid it, but to avoid it we must properly
      // maintain the frame-line value independently on the server... which
      // is tricky so for now we do it the safe way
      if (line == -1)
      {
         // couldn't find it locally, go down to the client to get it
         if (isRedirected() || !_isVisible())
            line = 0;
         else
            line =
               LogicalTerminal.getFrameLine(this, pendingUpDown).intValue();
      }
      
      return line;
   }
   
   /**
    * Processes any aggregator elements in the given array of frame elements.
    * The first time this method is invoked within a particular display loop,
    * an <code>AggregateHandler</code> instance is created.  This instance is
    * stored in an instance member, which is cleared when the display loop is
    * terminated.  Each time this method is invoked, the handler is requested
    * to accumulate data in all of the associated <code>Accumulator</code>
    * instances (if any) associated with the given frame elements.
    *
    * @param    fe
    *           The list of widgets that may need accumulation.
    * @param    accumulate
    *           Flag indicating if accumulation should be performed. If false,
    *           the widgets which have aggregators will be registered with
    *           this frame.
    */
   private void processAggregates(FrameElement[] fe, boolean accumulate)
   {
      // make a set of the accumulator widgets
      Set<FrameElement> wids = null;

      int len = fe.length;

      for (int i = 0; i < len; i++)
      {
         if (fe[i] != null && fe[i].statsColumn())
         {
            if (wids == null)
            {
               wids = new HashSet<FrameElement>();
            }
            wids.add(fe[i]);
         }
      }

      if (wids != null)
      {
         // create or update the handler
         AggregatorElement[] accums = wids.toArray(new AggregatorElement[0]);

         try
         {
            if (aggregateHandler == null)
               aggregateHandler = new AggregateHandler(accums);
            else
               aggregateHandler.verifyAccumulators(accums);
         }
         catch (IllegalStateException exc)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.warning(exc.getMessage());
            }
         }
      }
      
      // accumulate data
      if (accumulate && aggregateHandler != null)
      {
         aggregateHandler.accumulate(fe);
      }
   }
   
   /**
    * Reports if this frame is a page element (a header or footer). This is
    * only <code>true</code> in the case that the output stream is paged AND
    * this frame is marked as a page-top (header) or a page-bottom (footer).
    *
    * @param    out
    *           The output destination.
    *
    * @return   <code>true</code> if this is a header or footer.
    */
   boolean isPageElement(Stream out)
   {
      if (out == null)
         return false;
      
      if (out instanceof StreamWrapper)
      {
         StreamWrapper sw = (StreamWrapper) out;
         
         if (sw.dereference() == null)
            return false;
      }
      
      return (out.rawPageSize() > 0 && (getPageBottom() || getPageTop()));
   }
   
   /**
    * Reports if the unnamed stream is redirected at this time.
    *
    * @return   <code>true</code> if the unnamed stream is redirected
    */
   boolean isRedirected()
   {
      return UnnamedStreams.isOut();
   }

   /**
    * Reports if there is a cached output stream for use with redirecting
    * the output from aggregation.
    *
    * @return   <code>true</code> if aggregation output is redirected
    */
   boolean isOutRedirected()
   {
      return redirOut != null && redirOut.isOut();
   }
   
   /**
    * Check if output stream is a terminal.
    * 
    * @return   <code>true</code> if output stream is terminal.
    */
   boolean isOutTerm()
   {
      return redirOut != null && redirOut.isTerm();
   }
   
   /**
    * Full initialization processing. 
    *
    * @param    configClass
    *           The class of the configuration implementation class.
    * @param    configInstance
    *           The instance of the configuration implementation class.
    * @param    frameName
    *           The frame name.
    */
   private void initialize(Class<?> configClass,
                           Object configInstance, 
                           String frameName)
   {
      // register frame with LogicalTerminal
      frameId = LogicalTerminal.registerFrame(this);
      
      coreInitialize(configClass, configInstance, frameName);
      
      // create the associated screen buffer which is initially empty
      frameBuf = new ScreenBuffer(frameId, n2w.size());
   }
   
   /**
    * Core initialization processing. 
    *
    * @param    configClass
    *           The class of the configuration implementation class.
    * @param    configInstance
    *           The instance of the configuration implementation class.
    * @param    frameName
    *           The frame name.
    */
   private void coreInitialize(Class<?> configClass,
                               Object configInstance, 
                               String frameName)
   {
      this.configClass    = configClass;
      this.configInstance = configInstance;
      this.frameName      = frameName;
      
      WidgetList wl = (WidgetList) configInstance;
 
      // cache the widget map for this frame
      n2w = wl.getWidgets();
      
      // cache the widget name map for this frame
      n2n = wl.getNames(); 

      // cache the header widget map for this frame
      n2h = wl.getHeaderWidgets(); 
      
      // init the widget array
      widgets = new GenericWidget[n2w.size()];
      
      int wid = frameId;
      frame = new FrameWidget();
      frame.setName(this.frameName + "##" + this.configClass.getName());
      frame.setFrame(this);

      // add all the widgets in the order of their definition
      Iterator<String> iter = n2w.keySet().iterator();
      
      while (iter.hasNext())
      {
         // calculate the next widget ID
         wid++;
         
         // calculate the array index (we could hard code this to be slightly
         // faster, but it is better to make the relationship explicit
         int idx = (wid - (frameId + 1)); 
         
         // get the widget
         widgets[idx] = n2w.get(iter.next());
         
         // initialize key state in the widget
         widgets[idx].setFrame(this);
         widgets[idx].setId(wid);
      }

      // insert additional skip if headers are not empty
      if (n2h.size() > 0 && n2w.size() > 0)
      {
         SkipEntity skip = new SkipEntity(true);
         skip.setHeader(true);
         n2h.put("", skip);
      }
      
      // assign ids to header widgets in the order of their definition
      iter = n2h.keySet().iterator();
      
      while (iter.hasNext())
      {
         // calculate the next widget ID
         wid++;
         
         // get the widget
         GenericWidget gw = n2h.get(iter.next());
         
         // initialize key state in the widget
         gw.setFrame(this);
         gw.setId(wid);
         
         // we don't add these to the widgets[] since header widgets are not
         // real
      }

      // initialize the default field group
      getFirstChild();
      fieldGroup.setFrame(this);
   }
   
   /**
    * Drive the input stream reading process, where each frame element in the
    * list resolves to the read of a field from the stream, a bypass of the
    * field in the stream or an embedded assignment. This is an alternative
    * process to the traditional {@link #promptFor} processing.  In 
    * particular, while the screen buffer is modified during this processing,
    * the state of the UI is not otherwise modified.  This is just a way to  
    * redirect input from a stream and read that input from user interface
    * specific APIs.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    *
    * @param    in
    *           The input stream from which to read.
    * @param    wids
    *           The variables and widgets to process. Any <code>null</code>
    *           elements and widget only elements are ignored.
    * @param    assign
    *           <code>true</code> if the read data should be assigned back to
    *           the caller's <code>data</code> instance.
    *           
    * @throws   EndConditionException 
    * @throws   ErrorConditionException 
    */
   @SuppressWarnings("null")
   private void processStreamInput(Stream          in,
                                   FrameElement[]  wids,
                                   boolean         assign)
   throws EndConditionException,
          ErrorConditionException
   {
      if (in != null)
      {
         boolean abort = false;
         
         // errors are always doubled (see viewWorker()), we don't know why
         // Progress does this, but we duplicate it here
         try
         {
            abort = in.protectReads();
            
            if (abort)
            {
               // this only can occur in silent error mode (otherwise an
               // exception would have been thrown) so there are no errors to
               // double, just return
               return;
            }
         }
         
         catch (ErrorConditionException ece)
         {
            // ignore and mark the flag to allow the 2nd error to be generated
            abort = true;
         }
         
         // implement the second error if needed
         if (abort && in.protectReads())
            return;
      }
         
      in.resetCurrentLine();
      
      try
      {
         driver:
         for (int i = 0; i < wids.length; i++)
         {
            if (wids[i] != null && !wids[i].widgetOnly())
            {
               if (wids[i].assignOnly())
               {
                  if (assign)
                  {
                     // implement embedded assignments
                     wids[i].set(null);
                  }
               }
               else
               {
                  // honor bypass (the CARET in Progress)
                  Accessor access = (wids[i].bypass() ? null : wids[i]);
                  GenericWidget ww = wids[i].bypass() ? null 
                                                      : wids[i].getWidget();
                                                      
                  if (!getField(in, access, frameBuf, ww, null, assign))
                  {
                     break driver;
                  }
               }
            }
         }
         
      }
      
      finally
      {
         in.resetCurrentLine();
         
         Stream out = UnnamedStreams.output();
         
         // only echo if the input stream is marked for echoing AND if
         // no output redirection is active OR output redirection is active
         // AND that stream is marked for echoing
         if (in.getEcho() && (out == null || out.getEcho()))
         {         
            // cause echo of our data (whatever made it to the screen-buffer)
            view();
         }

         // set LASTKEY to 13 as if there was full line
         KeyReader.setLastKey(13);
      }
   }

   /**
    * Core widget accessor processing, which handles the various widget name
    * forms and the lookup of the widget including proper processing of array
    * elements.
    * <p>
    * This method handles name processing in both the J2SE-compatible form
    * as well as the "simplified" naming that always decapitalizes the first
    * letter of all widget names.
    * <p>
    * This method also handles the case when one widget represents entire 
    * array of widgets.
    *
    * @param    widgetName
    *           The base name of the widget to lookup, taken directly from
    *           the proxy method name.
    * @param    op
    *           The type of the operation (get, set or accessor).
    * @param    numArgs
    *           The number of arguments passed to the proxy handler. 
    * @param    index
    *           The index of an array element or <code>null</code> if no
    *           array processing is needed.
    *
    * @return   The widget by that name or <code>null</code> if no widget
    *           can be found by that name.
    */
   private GenericWidget getWidget(String widgetName,
                                   int    op,
                                   int    numArgs,
                                   Object index)   
   {
      // detect if this is an array form
      boolean array = (op == INVOKE_GET    && numArgs == 1) ||
                      (op == INVOKE_SET    && numArgs == 2) ||
                      (op == INVOKE_ACCESS && numArgs == 1);
                      
      // first try our "simplistic" but non-J2SE-compatible widget name which
      // always has the first char converted to lower case
      char[] cname = widgetName.toCharArray();

      if (cname.length > 0)
         cname[0] = Character.toLowerCase(cname[0]);
 
      String name = new String(cname);
      
      GenericWidget result = null; 
      
      if (array)    
      {
         // try non-indexed access
         result = getWidgetByName(name);
         
         if (result != null)
            return result;
         
         // for an indexed array element, recalculate the target widget name
         name = generateWidgetName(name, op, index);
      }
      
      result = getWidgetByName(name);

      if (result != null)
         return result;

      // the simplistic form failed, try the J2SE-compatible form
      name = Introspector.decapitalize(widgetName);
      
      if (array)  
      {
         // try non-indexed access
         result = getWidgetByName(name); 
         
         if (result != null)
            return result;
         
         name = generateWidgetName(name, op, index);
      }
      
      result = getWidgetByName(name);
      
      return result;
   }
   
   /**
    * Simple helper to lookup a widget by a given name.
    *
    * @param    name
    *           The case-sensitive widget name to retrieve.
    *
    * @return   The widget by that name or <code>null</code> if no widget
    *           can be found by that name.
    */
   private GenericWidget getWidgetByName(String name)
   {
      GenericWidget res = n2w.get(name);
      return (res != null) ? res : n2h.get(name);
   }
   
   /**
    * Returns the current screen value of the specified widget. 
    * 
    * @param    widgetName
    *           The widget name. 
    * @param    dataType
    *           The widget data type; a class representing one of the
    *           following widget data types:
    *           <ul>
    *             <li><code>character</code>;
    *             <li><code>integer</code>;
    *             <li><code>decimal</code>;
    *             <li><code>logical</code>;
    *             <li><code>date</code>.
    *           </ul>
    *
    * @return    <code>Object</code> currently in the screen buffer as the
    *            value of the widget or one of these defaults, if no value
    *            has been set:
    *           <ul>
    *             <li><code>character</code> - "";
    *             <li><code>integer</code> - 0;
    *             <li><code>decimal</code> - 0;
    *             <li><code>logical</code> - no (false);
    *             <li><code>date</code> - ? (unknown).
    *           </ul>
    */
   @SuppressWarnings("unused")
   private Object getter(String widgetName, Class<?> dataType)
   {
      int index = getWidgetId(widgetName);
      
      if (index < 0)
      {
         throw new RuntimeException("Invalid widget name " + widgetName);
      }
      
      return getter(index, dataType, true, false);
   }
   
   /**
    * Returns the current screen value of the specified widget. 
    * 
    * @param    widgetId
    *           The widget ID. 
    * @param    dataType
    *           The widget data type; a class representing one of the
    *           following widget data types:
    *           <ul>
    *             <li><code>character</code>;
    *             <li><code>integer</code>;
    *             <li><code>decimal</code>;
    *             <li><code>logical</code>;
    *             <li><code>date</code>.
    *           </ul>
    * @param    defaults
    *           <code>true</code> if a default value should be returned when
    *           the field is uninitialized in the screen buffer.
    * @param    validate 
    *           Perform validation.
    *
    * @return    <code>Object</code> currently in the screen buffer as the
    *            value of the widget or one of these defaults, if no value
    *            has been set:
    *           <ul>
    *             <li><code>character</code> - "";
    *             <li><code>integer</code> - 0;
    *             <li><code>decimal</code> - 0;
    *             <li><code>logical</code> - no (false);
    *             <li><code>date</code> - ? (unknown).
    *           </ul>
    */
   Object getter(int widgetId, 
                 Class<?> dataType, 
                 boolean defaults, 
                 boolean validate)
   {  
      // get the value from the screen buffer by index
      Object value = frameBuf.getWidgetValue(widgetId);
      GenericWidget widget = getWidgetForId(widgetId);
      String format = widget.getFormat();
      
      value = convertScreenValue(dataType, value, format);

      if (value == null)
      {
         // cut off early if value is null and no need to generate defaults
         if (!defaults)
            return null;
            
         // generate a default value if nothing is in the screen buffer
         if (BaseDataType.class.isAssignableFrom(dataType))
         {
            value = BaseDataType.generateDefault(dataType);
         }
         else
         {
            throw new RuntimeException("unsupported data type " +  dataType);
         }
      }
      
      if (validate && !formatFilter.contains(widgetId))
      {
         // finally perform check
         ErrorEntry res = DisplayFormatFactory.checkFormat((BaseDataType) value, 
                                                           format);
               
         if (res != null)
         {
            try
            {
               // if NO-ERROR is active or there are too many messages, 
               // then postpone error messaging and force final format check
               if (!ErrorManager.isSilent() && !silentFormatErrors)
               {
                  LogicalTerminal.message(res, 
                                          new ErrorEntry(71, 
                                                         "The value input for " + 
                                                         widget.getLabelStr() + 
                                                         " is not valid"));
                  silentFormatErrors = true;
               }
               else
               {
                  enableFinalFormatCheck();
               }
            }
            catch (Throwable e)
            {
               if (e instanceof StopConditionException)
                  throw (ConditionException) e;

               if (e instanceof QuitConditionException)
                  throw (ConditionException) e;

               // ignore everything else
            }
         }
      }
      
      return value;
   }

   /**
    * Enable format check at the end of input loop.  
    */
   private void enableFinalFormatCheck()
   {
      finalCheck = true;
   }

   /**
    * Convert given value into proper form for further conversion onto
    * SCREEN-VALUE.
    * <p>
    * In cases when <code>dataType</code> is {@link character} and the 
    * <code>value</code> is a logical value, then the result will be the 
    * logical value formatted using the given <code>format</code>.
    * 
    * 
    * @param   dataType
    *          Desired data type.
    * @param   value
    *          Source value to convert from.
    * @param   format
    *          The widget's format. Currently used only for logical values.
    * @return  Converted value.
    */
   private Object convertScreenValue(Class<?> dataType,
                                     Object value,
                                     String format)
   {
      if (value != null)
      {
         // handle type conversion if needed
         
         BaseDataType bdt = null;
         
         if (value instanceof FormattedValue)
         {
            // these cases occur when an @ base field is used and there is
            // a call to the getter before any editing occurs on the client
            // OR an edit on the client was done but no changes were made 
            FormattedValue fv = (FormattedValue) value; 
            
            bdt   = fv.getValue(); 
         }
         else if (value instanceof BaseDataType &&
                  !dataType.isInstance(value))
         {
            // some use of @ base fields may not use the FormattedValue but
            // the type conversion is still needed
            bdt = (BaseDataType) value;
         }
         
         if (bdt != null)
         {
            value = null;
            
            if (dataType.isInstance(bdt)) 
            {
               // the types are the same so we can simply return the value
               value = bdt;
            }
            else
            {
               // hmm, we need to convert this into the proper type
               if (dataType.equals(character.class))
               {
                  if (bdt.getClass().equals(logical.class))
                  {
                     // in this case, use the formatted representation
                     value = character.valueOf(bdt, format);
                  }
                  else
                  {
                     // use the common/safe way to convert from any BDT to
                     // the character type
                     value = character.valueOf(bdt);
                  }
               }
               else
               {
                  if (bdt instanceof unknown)
                  {
                     value = BaseDataType.generateUnknown(dataType);
                  }
                  else
                  {
                     // try to find a constructor that matches our type
                     Class<?>[] sig = new Class[] { bdt.getClass() };
                     
                     try
                     {
                        Constructor<?> cvt = dataType.getConstructor(sig);
                        
                        // best efforts to convert
                        if (cvt != null)
                        {
                           value = cvt.newInstance(new Object[] { bdt });
                        }
                     }
                     
                     catch (Exception exc)
                     {
                        // security problems, no such method... force default
                        // path below
                     }
                  }
               }
            }
         }
      }
      return value;
   }
   
   /**
    * Sets a new value for the specified widget in the screen buffer.
    * 
    * @param   widgetName
    *          widget name
    * @param   value
    *          An object representing the widget value. Since this code is
    *          called from the Java Reflection API, any time a primitive
    *          (e.g. <code>int</code>) is passed, Reflection automatically
    *          wraps it in an <code>Object</code> of the correct type (e.g.
    *          <code>Integer</code>).  Thus it is safe to always expect the
    *          value to be an <code>Object</code>.
    * @return  <code>null</code> to comply with <code>void</code>
    */
   @SuppressWarnings("unused")
   private Object setter(String widgetName, Object value)
   {
      int index = getWidgetId(widgetName);

      if (index < 0)
      {
         throw new RuntimeException("Invalid widget name " + widgetName);
      }      
      
      return setter(index, value);
   }
      
   /**
    * Sets a new value for the specified widget in the screen buffer.
    * 
    * @param   index
    *          The widget ID.
    * @param   value
    *          An object representing the widget value. Since this code is
    *          called from the Java Reflection API, any time a primitive
    *          (e.g. <code>int</code>) is passed, Reflection automatically
    *          wraps it in an <code>Object</code> of the correct type (e.g.
    *          <code>Integer</code>).  Thus it is safe to always expect the
    *          value to be an <code>Object</code>.
    * @return  <code>null</code> to comply with <code>void</code>
    */
   private Object setter(int index, Object value)
   {
      if (index < 0)
      {
         throw new RuntimeException("Invalid widget!");
      }      
      
      Object val = null;

      // wrapper the value according to these rules:
      // String  -> character
      // Integer -> integer
      // Double  -> decimal
      // Boolean -> logical
      if (value instanceof String)
      {
         val = new character((String) value);
      }
      else if (value instanceof Integer)
      {
         val = new integer(((Integer) value).intValue());
      }
      else if (value instanceof Double)
      {
         val = new decimal(((Double) value).doubleValue());
      }
      else if (value instanceof Boolean)
      {
         val = new logical(((Boolean) value).booleanValue());
      }
      else if (value instanceof BaseDataType)
      {
         // we need to make a copy of the original value because otherwise
         // any assign() in the calling code will actually modify the value
         // in the screen buffer, which would be incorrect behavior
         val = ((BaseDataType) value).duplicate();
         
         if (value instanceof unknown)
         {
            ControlEntity widget = (ControlEntity) getWidgetForId(index);
            val = translateUnknown((BaseDataType) value, widget);
         }
      }
      else
      {
         // basic Object support
         val = value;
      }
         
      // setting the value in the screen buffer by index
      frameBuf.putWidgetValue(index, val);

      return null;
   }
   
   /**
    * Generate an indexed widget name from a simple base name and a numeric
    * object that represents the index.  This method converts the numeric
    * object into a string using simple integer based formatting and then
    * concatenates this to the base name.
    *
    * @param    base
    *           The widget's base name.
    * @param    op
    *           The type of the method name (getter/setter/accessor).  Only
    *           used for error text.
    * @param    index
    *           The number to convert into a string representation.
    *
    * @return   The resulting name.
    *
    * @throws   RuntimeException
    *           If the index object is not a supported numeric object.
    */
   private String generateWidgetName(String base, int op, Object index)
   throws RuntimeException
   {
      String result = indexAsString(index);
      
      if (result == null)
      {
         String type = "accessor";
         
         if (op == INVOKE_GET)
         {
            type = "getter";
         }
         else if (op == INVOKE_SET)
         {
            type = "setter";
         }
         
         throw new RuntimeException("illegal " + type + " call");
      }
      
      StringBuffer sb = new StringBuffer(base).append(result.trim());            
      
      return sb.toString();
   }
   
   /**
    * Takes a numeric type (<code>NumberType</code>, <code>Integer</code> or
    * <code>Double</code>) and converts this into a string that can be used
    * as a name component for a widget getter/setter/accessor.
    *
    * @param    index
    *           The number to convert into a string representation.
    *
    * @return   The resulting string or <code>null</code> if the type is not
    *           one of the supported numeric types.
    */
   private String indexAsString(Object index)
   {
      String result = null;
      
      if (NumberType.class.isInstance(index))
      {
         result = ((NumberType)index).toString(">>>>>>>9");
      }
      else if (Integer.class.isInstance(index))
      {
         result = ((Integer)index).toString();
      }
      else if (Double.class.isInstance(index))
      {
         int idx = ((Double)index).intValue();
         
         result = Integer.toString(idx); 
      }
      
      return result;
   }
   
   /**
    * Ensure that any instance of the <code>unknown literal</code> is
    * converted into the proper type as expected by the given widget. 
    * <p>
    * The <code>unknown literal</code> is represented by the class
    * {@link unknown} which is a subclass of {@link BaseDataType}.  
    * This is only used to represent the literal and it needs to be converted 
    * to the proper type for use by the widget.
    * <p>
    * This must not be confused with a {@link BaseDataType} such as
    * an {@link integer} which is set to the <code>unknown value</code>.
    * Such cases are not literals.  Note that this method's purpose is to
    * convert the literal form into the proper non-literal type.
    *
    * @param    value
    *           The value to be converted if necessary.
    * @param    widget
    *           The widget being processed.
    *
    * @return   The original value if the type is not <code>unknown</code> or
    *           an instance of the proper type (set as 
    *           <code>unknown value</code>) otherwise.
    */
   private BaseDataType translateUnknown(BaseDataType  value,
                                         ControlEntity widget)
   {
      if (value instanceof unknown)
      {
         // honor the type of the widget's data
         Class<?> type = widget.getDataClass();
         
         // honor user attempt to override "unknown" date field behavior 
         // (empty string) 
         if (type == date.class)
            type = character.class;
         
         // get an instance that represents unknown value
         value = BaseDataType.generateUnknown(type);
      }
      
      return value;
   }

   /**
    * Set each value in the screen buffer for each widget using the widget
    * id as the index.  Please note that assign only, widget only and 
    * <code>null</code> entries in the array are ignored.
    *
    * @param    data
    *           The list of elements to copy into the screen buffer.
    */
   private void copyToScreenBuffer(FrameElement[] data)
   {
      for (int i = 0; i < data.length; i++)
      {  
         if ( data[i] != null      && 
             !data[i].assignOnly() &&
             !data[i].widgetOnly() &&
             !data[i].bypass())
         {
            Object value = (data[i].hasFormat()) 
                           ? (Object) data[i].getFormattedValue()
                           : (Object) data[i].get();
                           
            ControlEntity widget = (ControlEntity) data[i].getWidget();
            
            if (value instanceof unknown)
            {
               value = translateUnknown((BaseDataType) value, widget);
            }
            
            frameBuf.putWidgetValue(widget.getId(), value);
         }
      }
      
      if (data.length > 0 && triggerActive > 0)
      {
         LogicalTerminal.setHardFormatCheck(this.getFrameId());
      }
   }

   /**
    * Set value of each static widget mentioned in the list.
    * 
    * @param    data
    *           The list of elements.
    */
   private void setStaticWidgets(FrameElement[] data)
   {
      for (int i = 0; i < data.length; i++)
      {  
         if (data[i] == null                             || 
             data[i].assignOnly()                        ||
             data[i].widgetOnly()                        ||
             data[i].getWidget() instanceof FillInWidget ||
             !(data[i].getWidget() instanceof ControlTextWidget))
         {
            continue;
         }

         BaseDataType  value  = data[i].get(); 
         ControlEntity widget = (ControlEntity) data[i].getWidget();
         
         if (value instanceof unknown)
         {
            value = translateUnknown(value, widget);
         }
         
         // assuming one cannot use the @ base field feature with a static
         // widget, we shouldn't need to process format strings here so we 
         // keep this simple
         frameBuf.putWidgetValue(widget.getId(), value);
      }
   }
   
   /**
    * This method should be called anytime the frame buffer is sent down
    * to the thin client.
    * <p>
    * The latest header values will be obtained and stored in the header map.
    * Any subsequent display on the client will reflect the latest headers.
    * <p>
    * Any dynamic down value that has been registered will be queried and
    * stored in the frame buffer at this time.  The user of the frame may
    * have changed this value at any time.  We read the latest value from
    * the accessor (the delegated access to this variable) and if it is
    * positive it is stored in the screen buffer directly.  Any non-positive
    * value is stored as 0.  If there is no registered dynamic down value,
    * then this processing is bypassed completely.
    * <p>
    * Any dynamic title expression that has been registered will be evaluated
    * and the result will be stored in the frame buffer. If there is no 
    * registered dynamic title expression, then this processing is bypassed
    * completely.
    * <p>
    * There is no counterpart to this method upon return from the client 
    * since headers and dynamic down values can never be edited, they
    * are only sent down to the client (one way rather than bidirectional).
    */
   private void updateDynamicContent()
   {
      if (headers != null)
      {
         for (int i = 0; i < headers.length; i++)
         {
            frameBuf.putHeaderValue(headers[i].getWidget().getId(),
                                    headers[i].get());
         }
      }
      
      if (dynDown != null)
      {
         NumberType num  = (NumberType) dynDown.resolve();
         int        down = num.intValue();
         
         frameBuf.putDown((down < 0) ? 0 : down);
      }
      
      if (dynColumn != null)
      {
         NumberType num    = (NumberType) dynColumn.resolve();
         int        column = num.intValue() - 1;
         
         frameBuf.putColumn((column < 0) ? 0 : column);
      }

      if (dynRow != null)
      {
         NumberType num = (NumberType) dynRow.resolve();
         int        row = num.intValue() - 1;
         
         frameBuf.putRow((row < 0) ? 0 : row);
      }
      
      updateDynamicTitle();
      updateDynamicBrowseTitle();
   }
   
   /**
    * Update dynamic browse title.
    */
   private void updateDynamicBrowseTitle()
   {
      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i] instanceof BrowseWidget)
         {
            BrowseWidget brws = (BrowseWidget) widgets[i];
            
            brws.updateDynamicTitle();
         }
      }
   }

   /**
    * Update dynamic frame title.
    */
   protected void updateDynamicTitle()
   {
      if (dynTitle != null)
      {
         character title = null;
      
         if (frameBuf.getTitle() != null)
         {
            // frame already has title, so calculate it silently
            // Note that we can't just skip title calculation here
            // because frame can be hidden/viewed at client during
            // normal client side frame processing.
            boolean quiet = ErrorManager.isIgnore();
            
            if (!quiet)
            {
               ErrorManager.setIgnore(true);
            }
            
            try
            {
               title = dynTitle.execute();
            }
            
            finally
            {
               if (!quiet)
               {
                  ErrorManager.setIgnore(false);
               }
            }
         }
         else
         {
            // Bracket the dynamic title expression execute() in silent error
            // mode if not already enabled.
            boolean warning = ErrorManager.isWarningMode();
            
            if (!warning)
            {
               ErrorManager.warningModeEnable();
            }
            
            try
            {
               title = dynTitle.execute();
            }
            
            finally
            {
               if (!warning)
               {
                  ErrorManager.warningModeDisable();
               }
            }
         }
         
         String txt = (title.isUnknown() ? "" : title.toStringMessage());
         
         // force the title to go down with the next screen buffer
         frameBuf.putTitle(txt);
         
         // keep the screen definition in sync BUT don't trigger a push here
         FrameConfig fc = (FrameConfig)frame.getConfig();
         fc.setTitle(txt);
      }      
   }

   /**
    * Process list of frame elements and link widgets into text (word wrap)
    * groups if necessary. Resulting set of widgets is returned as list for
    * use by other methods.
    * 
    * @param   list
    *          Array of <code>FrameElement</code> instances.
    * 
    * @return  List of frame elements which refer widgets linked into text
    *          groups. All frame elements are placed in one list regardless
    *          from group.
    */
   private List<FrameElement> textGroup(FrameElement[] list)
   {
      ArrayList<FrameElement> res = new ArrayList<FrameElement>();
      
      FillInWidget prev = null;
      int currGroup = -1;
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] == null)
            continue;
         if (list[i].wordWrap())
         {
            res.add(list[i]);
            
            int id = ((WordWrapElement) list[i]).getIndex();

            if (id != currGroup)
               prev = null;
            
            currGroup = id;
            
            FillInWidget fin = (FillInWidget) list[i].getWidget();
            fin.startGrouping(prev);
            prev = fin;
         }
      }
      
      if (res.size() > 0)
         pushScreenDefinition();
      
      return res;
   }

   /**
    * Reset text grouping (word wrap mode).
    * 
    * @param   list
    *          List of <code>FrameElement</code> instances.
    */
   private void textUngroup(List<FrameElement> list)
   {
      for (Iterator<FrameElement> iter = list.iterator(); iter.hasNext();)
      {
         FrameElement item = iter.next();
         
         ((FillInWidget) item.getWidget()).resetGrouping();
      }
      
      if (list.size() > 0)
         pushScreenDefinition();
   }
   
   /**
    * Get the value in the screen buffer for each widget using the widget
    * id as the index, then assign this back to the variable from the 
    * caller.  Please note that widget only and <code>null</code> entries
    * in the array are ignored.
    * <p>
    * This method honors {@link EmbeddedAssignment} elements in the proper
    * order in the array.
    *
    * @param    data
    *           The list of elements to copy from the screen buffer.
    * @param    defaults
    *           <code>true</code> means that default values will be used
    *           if screen buffer does not contain updated values.
    */
   private void copyFromScreenBuffer(FrameElement[] data, boolean defaults)
   {
      // bracket the assignment with a notification
      TransactionManager.batchStart();
      
      for (int i = 0; i < data.length; i++)
      {  
         if (data[i] != null && !data[i].widgetOnly() && !data[i].bypass())
         {
            BaseDataType value = null; 
            
            // read from screen buffer if we are not assign only
            if (!data[i].assignOnly())
            {
               GenericWidget wid  = data[i].getWidget();
               Class<?>      type = ((ControlEntity) wid).getDataClass();
               
               // get the value out of the screen buffer
               value = (BaseDataType) getter(wid.getId(), 
                                             type, 
                                             defaults, 
                                             false);                              
               
               if (value == null)
               {
                  // no value in the screen buffer, so don't assign back
                  continue;
               }
            }
            
            // assign the value back (embedded assignments handle their
            // own data value, so we pass null in these cases)
            data[i].set(value);
         }
      }      
      
      // end of assign bracket
      TransactionManager.batchStop();      
   }
   
   /**
    * Makes the frame and all its field-level widgets visible.
    */
   private void makeFrameVisible()
   {
      // make all field-level widgets visible, unless are marked as HIDDEN
      for (int i = 0; i < widgets.length; i++)
      {
         if (!widgets[i].getConfig().isHidden())
         {
            widgets[i].getConfig().setVisible(true);
         
            if (widgets[i] instanceof BrowseWidget)
               ((BrowseWidget) widgets[i]).finishSetup();
         }
      }

      // make the frame visible
      ((FrameConfig)frame.getConfig()).setVisible(true);
      // set the hidden flag to OFF
      ((FrameConfig)frame.getConfig()).setHidden(false);
   }
   
   /**
    * Create an array of all widgets in this frame that are in the given
    * array.
    *
    * @param    list
    *           The widgets to include in the resulting list.
    *
    * @return   The list of remaining ids or <code>null</code> if no ids
    *           were included.
    */
   public int[] validIds(GenericWidget[] list)  
   {
      if (list == null || list.length == 0)
         return null;

      Set<Integer> widgetIds = new LinkedHashSet<Integer>();

      // add included widget IDs
      for (int i = 0; i < list.length; i++)
      {
         // null values are silently ignored
         if (list[i] == null)
            continue;
         
         int id = getWidgetId(list[i]);
         
         // invalid widget
         if (id == -1)
            continue;
         
         widgetIds.add(new Integer(id));
      }
     
      return idSetToArray(widgetIds);
   }

   /**
    * Makes widgets in the frame invisible.
    * 
    * @param    data
    *           Array of widgets to hide.
    */
   @Override
   public void hide(FrameElement[] data)
   {
      hide(data, false);
   }

   /**
    * Makes widgets in the frame invisible.
    * 
    * @param    data
    *           Array of widgets to hide.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(FrameElement[] data, boolean noPause)
   {
      //TODO: add batch hide to ClientExports and use it
      for (FrameElement frameElement : data)
      {
         frameElement.getWidget().hide(noPause);
      }
   }
   
   /**
    * Refresh current DOWN value.
    * 
    * @param    down
    *           New DOWN value reported by client.
    */
   void refreshDown(int down)
   {
      dynamicDown = down;
   }

   /**
    * Reset 'flushed' flag. This means that frame need to be flushed manually 
    * if there is an unexpected exit from loop. 
    */
   void resetFlushed()
   {
      flushed = false;
   }
   
   /**
    * Return <code>true</code> if frame content was flushed since last VIEW.
    * 
    * @return   <code>true</code> if frame was flushed.
    */
   boolean isFlushed()
   {
      return flushed;
   }
   
   /**
    * Create an array of all widgets in this frame that are NOT in the 
    * given array.
    *
    * @param    xcpt
    *           The widgets to exclude from the resulting list.
    *
    * @return   The list of remaining ids or <code>null</code> if all ids
    *           were excluded.
    */
   private int[] remainingIds(GenericWidget[] xcpt)
   {
      // create a set of all widget IDs
      Set<Integer> widgetIds = new LinkedHashSet<Integer>();
      
      int base = frameId + 1;
      int end  = base + widgets.length;
      
      for (int j = base; j < end; j++)
      {
         widgetIds.add(new Integer(j));
      }
      
      // remove excluded widget IDs
      for (int i = 0; i < xcpt.length; i ++)
      {
         // null values are silently ignored
         if (xcpt[i] == null)
            continue;
         
         int id = getWidgetId(xcpt[i]);
         
         // invalid widget
         if (id == -1)
            continue;
         
         widgetIds.remove(new Integer(id));
      }
      
      return idSetToArray(widgetIds);
   }
   
   /**
    * Converts the given set of <code>Integer</code> values into an array
    * of primitive <code>int</code>.
    *
    * @param    widgetIds
    *           The set of ids to convert.
    *
    * @return   The <code>int[]</code> or <code>null</code> if there are no
    *           ids in the set.
    */
   private int[] idSetToArray(Set<Integer> widgetIds)
   {  
      // calculate the widgets
      int size = widgetIds.size();
      
      if (size == 0)
      {
         return null;
      }

      int[] ids = new int[size];
      Integer[] wraps = widgetIds.toArray(new Integer[0]);

      for (int i = 0; i < size; i ++)
      {
         ids[i] = wraps[i].intValue();
      }
      
      return ids;
   }
   
   /**
    * Converts a widget array into an array of widget IDs.
    *
    * @param    widgets
    *           array of widgets which is to be edited.
    *
    * @return   The array of widget IDs.
    */
   private int[] idsFromWidgets(GenericWidget[] widgets)
   {
      if (widgets == null)
         return null;

      int[] ids = new int[widgets.length];

      for (int i = 0; i < widgets.length; i ++)
      {
         ids[i] = getWidgetId(widgets[i]);
      }
      
      return ids;
   }

   /**
    * Converts a frame element array into a array of widgets.  Any nodes that
    * are assignment only result in a <code>null</code> entry in the array.
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    *
    * @return   The array of widgets.
    */
   private GenericWidget[] widgetListFromFrameElements(FrameElement[] widgets)
   {
      GenericWidget   current = null;
      GenericWidget[] wids    = new GenericWidget[widgets.length];
      
      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i] == null || widgets[i].assignOnly())
         {
            current = null;
         }
         else
         {
            current = widgets[i].getWidget();
         }
         
         wids[i] = current;
      }
      
      return wids;
   }
   
   /**
    * Raise an error condition with the given error number and text IF
    * unnamed input has been redirected.  If silent error mode is enabled,
    * any error condition will be logged and this method will return.
    *
    * @param    num
    *           Error number.
    * @param    txt
    *           Error text.
    *
    * @return   <code>true</code> if in silent error mode and an error was
    *           raise.  <code>false</code> if the unnamed input is not
    *           redirected.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   private boolean errorOnRedirectedInput(int num, String txt)
   {
      if (UnnamedStreams.input() != null)
      {
         ErrorManager.recordOrThrowError(num, txt);
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Executes delayed conditional cursor movement.
    */
   private void conditionalUpDown()
   {
      if (pendingUpDown != 0)
      {
         int frameLine = frameLineWorker();
         
         LogicalTerminal.setConditionalUpDown(pendingUpDown, frameId);
         pendingUpDown = 0;
         
         if (frameLine < 0 || frameLine > dynamicDown)
         {
            // if after the conditional up/down we land outside the frame,
            // the cache must be reset
            screenBufferCache.clear();
         }
      }
      
      afterDown  = false;
      afterUnderline = false;
   }
   
   /**
    * DOWN worker routine for the unnamed stream, redirected or not.
    * 
    * @param    lines
    *           Number of lines to scroll down.
    *
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    */
   private void downWorker(int lines, boolean cond)
   {
      if (cond)
         lines = 1;

      if (aggregateHandler != null)
            aggregateHandler.notifyDown((cond) ? 0: lines);
      
      if (lines < 0 && isRedirected())
      {
         // this is a NOP
         return;
      }
      
      // for unnamed streams, we must end any PUT or EXPORT output
      int sid = (isRedirected() && !cond)
                ? getRemoteStreamId(UnnamedStreams.output())
                : -1;

      // save the current FRAME-LINE value to be used with generic down
      int savedFrameLine = frameLineWorker();

      if (pendingUpDown > 0 && afterUnderline)
         pendingUpDown--;
      
      int tmp = pendingUpDown;
      pendingUpDown = 0;

      LogicalTerminal.down(this, lines + tmp, cond, sid);
      afterDown = true;
      afterUnderline = false;
          
      if (isRedirected())
      {
         // redirected streams don't use the cache algorithm
         if (lines > 0) 
         {
            // reset the screen buffer if it is a down instruction
            frameBuf = new ScreenBuffer(frameId, n2w.size());
         }
         return;
      }
      
      // save down, retain and scroll values for this frame, to be used later
      FrameConfig frameConfig = ((FrameConfig) frame.getConfig());
      
      // avoid going to the client with frameDown() since it is expensive
      int down   = dynamicDown;
      int retain = frameConfig.getRetain();
      int scroll = frameConfig.getScroll();

      // check if both retain and scroll are zero
      boolean noRetainOrScroll = retain == 0 && scroll == 0;
      
      // check if the up/down is within the frame row-boundaries
      int     newFrameLine  = savedFrameLine + lines;
      boolean defaultUpDown = newFrameLine >= 1 && newFrameLine <= down;

      if (defaultUpDown)
      {
         // handle a generic up/down
         // cache the current buffer in the frame line as saved
         // before the up/down
         addScreenBufferToCache(this.frameBuf, savedFrameLine);

         // show the buffer associated to the frame line row
         // also make visible this screen buffer to the screen
         showScreenBuffer(frameLineWorker());                     
      }
      else
      {
         if (noRetainOrScroll)
         {
            // clear all row buffers
            this.screenBufferCache.clear();
            // cache the buffer for the current line
            showScreenBuffer(frameLineWorker());
            // all frame is reset, no additional display is needed
         }
         else 
         {    
            // data about the first part
            int start1 = 0;
            int end1   = 0;
            
            // data about second part
            int incr2  = 0;
            int start2 = 0;
            int end2   = 0;

            // data about third part
            int dir3  = 0;
            int end3  = 0;
            
            if (lines < 0)
            {
               // an up with retain or scroll
               start1 = 1;

               incr2  = -1;
               start2 = down;
               end2   = 1;

               if (retain != 0)
                  end1 = retain;
              
               if (scroll != 0)
                  end1 = down - scroll;
            }
            else
            {
               // a down with retain or scroll                  
               end1 = down;
               
               incr2  = 1;
               start2 = 1;
               end2   = down;

               if (retain != 0)
                  start1 = down - retain + 1;
               
               if (scroll != 0)
                  start1 = scroll + 1;
            }
            
            if (retain > 0)
            {
               dir3  = -1; 
               end3  = down - retain;                  
            }
            if (scroll > 0)
            {
               dir3  = 1; 
               end3  = scroll;                  
            }

            // save old cache
            Map<Integer, ScreenBuffer> oldCache = screenBufferCache;

            // new cache in which to store the new rows
            screenBufferCache = new HashMap<Integer, ScreenBuffer>();
            
            // first part
            
            for (int row = start1; row <= end1; row += 1)
            {
               int newKey = 0;
               
               // up
               if (lines < 0)
               {
                  if (retain != 0)
                     newKey = down - retain + row;
                  if (scroll != 0)
                     newKey = down - row + 1;
               }
               
               // down
               if (lines > 0)
               {
                  if (retain != 0)
                     newKey = row - (down - retain);
                  if (scroll != 0)
                     newKey = row - scroll;
               }
               
               ScreenBuffer oldBuf = oldCache.get(row);
               addScreenBufferToCache(oldBuf, newKey);
            }
            
            if (lines < 0)
            {
               addScreenBufferToCache(frameBuf, frameLineWorker());
            }
               
            // second part - to make the new rows visible, iterate 
            // through the cached rows
            for (int row = start2; 
                (incr2 > 0 && row <= end2) || (incr2 < 0 && row >= end2); 
                row += incr2)
            {
               showScreenBuffer(row);

               if (row != end2)
                  LogicalTerminal.down(this, incr2, cond);
            }
            
            // third part - position the cursor to the correct row
            for (int i = 1; i < end3; i++)
            {
               LogicalTerminal.down(this, dir3, cond);
            }                  

            // make the current screen buffer visible
            showScreenBuffer(frameLineWorker());
         }
      }
   }
   
   /**
    * This method will send the screen buffer for row at given index
    * to be displayed by the client, only if no redirected output occurred.
    * <p>
    * All widgets state must be set to changed so they will be updated. 
    * 
    * @param    idx
    *           The index of the line to be shown to the screen by the 
    *           client. The client MUST be positioned to the right row.
    */
   private void showScreenBuffer(int idx) 
   {
      if (isOutRedirected())
      {
         return;
      }

      ScreenBuffer buff = getScreenBufferFromCache(idx);
      this.frameBuf = buff;

      ((FrameConfig) frame.getConfig()).setModified(true);
      
      // set all widgets to changed
      Iterator<Integer> iter = buff.getWidgetIDIterator();
      ArrayList<Integer> ids = new ArrayList<Integer>();
      
      while (iter.hasNext()) 
      {
         Integer id = iter.next();
         ids.add(id);
         buff.setState(id, ScreenBuffer.CHANGED);
      }
      
      LogicalTerminal.view(this, Utils.integerCollectionToPrimitive(ids));
   }

   /**
    * DOWN worker routine for named streams.
    * <p>
    * For a named stream, UP statements have no effect on the FRAME-LINE 
    * index or on the screen buffer. DOWN statements will advance to the 
    * next row (even past the FRAME-DOWN row) and will reset the screen
    * buffer.
    * 
    * @param    lines
    *           Number of lines to scroll down.
    * @param    id
    *           The stream ID to which the terminal will be redirected during
    *           this processing or -1 if no temporary redirection is needed.
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    endOut                
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    */
   private void downWorker(int lines, int id, boolean cond, boolean endOut)
   {
      if (cond)
         lines = 1;
      
      if (aggregateHandler != null)
            aggregateHandler.notifyDown((cond) ? 0 : lines);

      if (lines < 0 && id != -1)
      {
         // this is a NOP
         return;
      }

      int tmp = pendingUpDown;
      pendingUpDown = 0;

      if (id == -1)
      {
         // just for safety
         LogicalTerminal.down(this, lines + tmp, cond);
      }
      else
      {
         // Same data for every iteration - bypass down on client side
         if (!sameViewDataRedirected)
            LogicalTerminal.down(this, (lines + tmp), id, cond, endOut);
         
         // For non-down report related frames we have to preserve current
         // screen buffer
         if (!frame.isDown() && isOutRedirected() && sameViewDataRedirected)
            return;

         if (lines > 0) 
         {
            // reset the screen buffer if it is a down instruction
            frameBuf = new ScreenBuffer(frameId, n2w.size());
         }
         
         // the UP instruction has no effect on the FRAME-LINE index
         // or on the buffer
      }
   }

   /**
    * UP worker routine.
    * 
    * @param    lines
    *           Number of lines to scroll up.
    */
   private void up(int lines)
   {
      downWorker(-lines, false);
   }
   
   /**
    * UP worker routine.
    * 
    * @param    lines
    *           Number of lines to scroll up.
    * @param    id
    *           The stream ID to which the terminal will be redirected during
    *           this processing or -1 if no temporary redirection is needed.
    */
   private void up(int lines, int id)
   {
      downWorker(-lines, id, false, false);
   }
   
   /**
    * Dereferences the given stream to find and return the remote stream ID.
    *
    * @param    out
    *           The stream (may be a <code>StreamWrapper</code> that has
    *           had a <code>RemoteStream</code> assigned OR it may be a
    *           <code>RemoteStream</code> instance).  If <code>null</code>
    *           or any wrapped reference is <code>null</code>, then the
    *           result will be treated as referencing the terminal and a
    *           -1 will be returned.
    *
    * @return   The remote stream ID.
    */
   int getRemoteStreamId(Stream out)
   {
      RemoteStream rs = null;
      
      if (out instanceof RemoteStream)
      {
         rs = (RemoteStream) out;
      }
      else if (out instanceof StreamWrapper)
      {
         StreamWrapper wrap = (StreamWrapper) out; 
         rs = (RemoteStream) wrap.dereference();
      }
      
      if (rs == null)
         return -1;

      // terminal stream with page size 0 is a regular interactive terminal
      // which does not require special handling
      if (rs.isTerm() && rs.getPageSize().intValue() == 0)
         return -1;
      
      return rs.getId();
   }
   
   /**
    * Adds a screen buffer instance to the cache for the specified row.
    * 
    * @param   buff
    *          The buffer to be added to the cache.
    * @param   row
    *          The 1-based row to which this buffer corresponds to.
    */
   private void addScreenBufferToCache(ScreenBuffer buff, int row) 
   {
      this.screenBufferCache.put(row, buff);
   }

   /**
    * Gets the screen buffer associated for this row. 
    * If there is no screen buffer associated with this row, then the method
    * will return a new empty buffer.
    * 
    * @param   row
    *          The 1-based row for which to get the current screen buffer.
    *          
    * @return  The buffer associated with this row.
    */
   private ScreenBuffer getScreenBufferFromCache(int row) 
   {
      ScreenBuffer buff = null;
      
      if (this.screenBufferCache.containsKey(row) == false) 
      {
         buff = new ScreenBuffer(frameId, n2w.size());

         // all static text widgets need to be copied from the previous buffer
         for (int i = 0; i < buff.size(); i++)
         {
            int wid = frameId + (i + 1);
            GenericWidget widget = this.getWidgetForId(wid);
            
            if (widget instanceof ControlTextWidget)
            {
               // if this is a static widget
               ControlTextWidget text = (ControlTextWidget) widget;
               ControlTextConfig config = (ControlTextConfig) text.getConfig();
               if (config.isStatic())
               {
                  buff.putWidgetValue(wid, frameBuf.getWidgetValue(wid));
               }
            }
         }
         
         addScreenBufferToCache(buff, row);
      }
      else 
      {
         buff = this.screenBufferCache.get(row);
      }
      
      return buff;
   }
   
   /**
    * Provides a cleanup mechanism for editing blocks that exits abnormally
    * (without a natural exit via {@link #continueEditing}).
    */
   private class EditingCleaner
   implements ConditionListener
   {
      /** List of widgets that need cleanup. */
      private FrameElement[] list = null;
      
      /** Track clean exit. */
      private boolean condition = false;
      
      /**
       * Create an instance, storing key data for cleanup purposes.
       * 
       * @param   list
       *          Array of frame elements. 
       */
      EditingCleaner(FrameElement[] list)
      {
         this.list = list;
         TransactionManager.registerFinalizable(this, false);
      }
      
      /**
       * Uses the finish notification provided by the {@link
       * TransactionManager} to ensure that any abnormal exit is detected
       * and cleanup occurs.
       */
      @Override
      public void finished()
      {
         boolean clean = false;
         
         synchronized (editLock)
         {
            if (currentOp != NONE)
            {
               clean = true;
            }
         }
         
         if (clean)
         {
            stopEditingMode(list, !condition, true);
         }
      }      
      
      /**
       * Processes the iterate notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      @Override
      public void iterate()
      {
      }      
      
      /**
       * Processes the retry notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      @Override
      public void retry()
      {
      }
      
      /**
       * Track all conditions which may break the editing mode.
       * 
       * @param   condition
       *          Unused.
       */
      @Override
      public void condition(Condition condition)
      {
         this.condition = true; 
      }            
   }

   /**
    * Manages <code>AggregatorElements</code> associated with the enclosing
    * frame instance.  An array of such elements is provided at construction
    * and is stored.  The constructor registers a new instance with the
    * <code>TransactionManager</code> at the nearest enclosing loop.  Each
    * invocation of the enclosing frame's <code>display</code> method causes
    * all associated accumulators to accumulate the data associated with that
    * pass of the loop.  At the bottom of each iteration of the loop, this
    * object checks whether any of its managed accumulators is currently at
    * the last row in a break group.  If so, sub-results are displayed.  At
    * the end of loop, final results of all accumulators which support such a
    * result are displayed.
    */
   private class AggregateHandler
   implements Finalizable
   {
      /** All aggregator elements handled by this object */
      private LinkedHashSet<AggregatorElement> accums = 
                                       new LinkedHashSet<AggregatorElement>();
      
      /** Elements which need to be displayed on finished processing */
      private List<AggregatorElement> fulls = 
                                       new ArrayList<AggregatorElement>();
      
      /** Elements which need to be displayed on next iteration */
      private List<AggregatorElement> subs = 
                                       new ArrayList<AggregatorElement>();
      
      /** Comparator used to sort AggregatorElements */
      private Comparator<AggregatorElement> comparator = 
                                       new AggregatorElementComparator();

      /** Number of lines in last DOWN call. */
      private int lastDown = 0;
      
      /**
       * If <code>true</code> then last operation was display() rather than
       * accumulate().
       */
      private boolean displayed = false; 
      
      /**
       * Construct a new instance, detecting which of the given aggregator
       * elements must produce a "final" result line at loop termination.
       * This objects registers itself with the transaction manager at the
       * nearest enclosing loop block.
       *
       * @param   accums
       *          All aggregator elements to be managed by this handler.
       *
       * @throws  IllegalStateException
       *          if no enclosing, looping block can be identified to which
       *          to register this <code>Finalizable</code>.
       */
      AggregateHandler(AggregatorElement[] accums)
      {
         verifyAccumulators(accums);
      }
      
      /**
       * Frame calls this method to notify handler about last down() call.
       * 
       * @param   lines
       *          Number of lines passed to down().
       */
      public void notifyDown(int lines)
      {
         lastDown  = lines;
      }
      
      /**
       * Compare provided array of accumulator elements with set of known ones
       * and add ones which are not yet known.
       * 
       * @param   elem
       *          Array of {@link AggregatorElement} instances to verify.
       */
      public void verifyAccumulators(AggregatorElement[] elem)
      {
         for (AggregatorElement next : elem)
         {
            if (accums.contains(next))
               continue;
            accums.add(next);
            
            if (!next.getAccumulator().isSubgroupOnly())
            {
               fulls.add(next);
            }
         }
         
         Collections.sort(fulls, comparator);
      }

      /**
       * Get the current value for all data being accumulated and cause each
       * registered accumulator to accumulate these data.  Detect and sort
       * the subset of aggregators (if any) which need to display their break
       * group results at the bottom of the current loop iteration.
       * <p>
       * If this handler is not managing any aggregators, this method returns
       * immediately.
       * 
       * @param   fe
       *          Array of frame elements.  
       */
      void accumulate(FrameElement[] fe)
      {
         HashMap<FrameElement, FrameElement> feMap = new HashMap<FrameElement, FrameElement>();
         for (int i = 0; i < fe.length; i ++)
         {
            feMap.put(fe[i], fe[i]);
         }

         for (AggregatorElement next : accums)
         {
            Accumulator accum  = next.getAccumulator();
                        
            if (feMap.containsKey(next))
            {
               FrameElement frameElem = feMap.get(next);
               BaseDataType datum = frameElem.get();
               accum.accumulate(datum);
            }
         }
         
         displayed = false;
      }
      
      /**
       * Display break group results (if any).  This method is invoked by the
       * <code>TransactionManager</code> at the top of the second and 
       * subsequent loop iteration.  This means that the handler has had a
       * chance to accumulate the current row's results.
       */
      @Override
      public void iterate()
      {
         subs.clear();
         
         for (AggregatorElement next : accums)
         {
            Accumulator accum  = next.getAccumulator();
                        
            if (accum.isLastOfGroup())
            {
               subs.add(next);
            }
         }
         
         if (!subs.isEmpty())
         {
            Collections.sort(subs, comparator);

            display(subs);
         }
         else
         {
            displayed = false;            
         }
      }
      
      /**
       * Processes the retry notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      @Override
      public void retry()
      {
      }      
      
      /**
       * Display full results (if any).  This method is invoked by the
       * <code>TransactionManager</code> after the display loop terminates,
       * after the handler has had a chance to accumulate the final row's
       * results.
       * <p>
       * This method resets the enclosing frame's aggregate handler reference.
       */
      @Override
      public void finished()
      {
         try
         {
            // force any pending subs out first (since iterate() is not
            // called at the bottom of the block)

            if (!fulls.isEmpty())
               display(fulls);
            else
               down();
         }
         finally
         {
            // reset aggregate handler
            aggregateHandler = null;
         }
      }

      /**
       * Convert list of {@link AggregatorElement} instances into array of
       * widget IDs.
       * 
       * @param   accums
       *          List of elements.
       * @param   handleFulls
       *          Flag which indicates processing of final results.
       * 
       * @return  The array of widget IDs.
       */
      private int[] getIDs(List<AggregatorElement> accums, boolean handleFulls)
      {
         LinkedHashSet<Integer> ids = new LinkedHashSet<Integer>(5);
         
         for (Element element : accums)
         {
            if (handleFulls && element.getAccumulator().isSubOnly())
               continue;
            
            ids.add(element.getWidget().getId());
         }

         return Utils.integerCollectionToPrimitive(ids);
      }
      
      /**
       * Display aggregation results, either for a single break group, or for
       * the entire set of rows displayed by a display loop.
       *
       * @param   accums
       *          List of aggregator elements to display.
       */
      private void display(List<AggregatorElement> accums)
      {
         boolean redirect = isOutRedirected();
         
         boolean handleFulls = accums == fulls;
         
         // check for accumulators which should 
         // produce output for entire data set
         if (handleFulls)
         {
            boolean produceOut = false;
            
            for (AggregatorElement elem : accums)
            {
               if (elem.getAccumulator().isSubOnly())
                  continue;
               
               produceOut = true;
               break;
            }
            
            if (!produceOut)
               return;
         }
         
         int tempOut = -1;
         
         if (redirect)
         {
            // temporarily switch the terminal to a new output destination
            int id = getRemoteStreamId(redirOut); 
            tempOut = LogicalTerminal.switchOutput(id, true, true);
         }
         
         try
         {
            if (displayed)
            {
               up();
               up();
            }
               
            if (lastDown > 0)
               up();
            
            // underline the columns being accumulated
            int len = accums.size();
            Iterator<AggregatorElement> iter;
            int[] ids = getIDs(accums, handleFulls);
            
            if (hadView == false)
            {
               List<AggregatorElement> wset = 
                  new ArrayList<AggregatorElement>(ids.length);
               Set<Integer> idSet = new HashSet<Integer>(ids.length);
               for (AggregatorElement elem : accums)
               {
                  if (idSet.contains(elem.getWidget().getId()) == false)
                  {
                     wset.add(elem);
                     idSet.add(elem.getWidget().getId());
                  }
               }
               view(wset.toArray(new AggregatorElement[0]));
               up();
            }
            underline(ids);
            
            ScreenBuffer sb = new ScreenBuffer(frameId, n2w.size());
            
            // process all columns with accumulation
            
            iter = accums.iterator();
            
            for (int i = 0; iter.hasNext(); i++)
            {
               AggregatorElement next = iter.next();
               GenericWidget wid = next.getWidget();
               Accumulator accum = next.getAccumulator();

               // skip accumulators which produce only subtotals
               if (handleFulls && accum.isSubOnly())
                  continue;
               
               String fmt   = next.hasFormat() ? next.getFormat() : null;
               Object value = (accums == subs) ? accum.getSubResult()
                                               : accum.getFullResult();
                                                  
               if (fmt != null)
               {
                  value = new FormattedValue((BaseDataType) value, fmt);
               }
               
               int[] widgetIds = new int[] { statsColumn.getId(), wid.getId()};
               
               // add the statistic to the screen buffer for the correct 
               // widget
               sb.putWidgetValue(wid.getId(), value);
               
               int weight = accum.getWeight();
               int nweight = (i == len - 1)
                             ? Integer.MAX_VALUE
                             : accums.get(i + 1).getAccumulator().getWeight();
               
               // this is the last set of outputs or a new set will begin on
               // the next iteration
 
               if (weight != nweight)
               {
                  // add the statistic label to the to the extra stats widget
                  character label = new character(accum.getLabel());
                  sb.putWidgetValue(statsColumn.getId(), label);
                  
                  // output the stats line
                  conditionalUpDown();

                  // the following call is qualified as DISPLAY in LogicalTerm
                  // note that we display exactly two widgets to avoid 
                  // triggering of repainting of all widgets in the frame
                  LogicalTerminal.view(GenericFrame.this, sb, widgetIds);
                  
                  // force a new line
                  
                  if (accums == subs)
                     down(1);
                  else
                     down();
   
                  // clear the screen buffer
                  sb.reset();
               }
            }
         }
         
         finally
         {
            if (redirect)
            {
               // put the old terminal back (either interactive or a
               // different redirection)
               LogicalTerminal.switchOutput(tempOut, false);
            }
            displayed = true;
         }
      }

      /**
       * A comparator implementation which compares
       * <code>AggregatorElement</code>s using only the weights of their
       * contained <code>Accumulator</code> instances.
       *
       * @see  com.goldencode.p2j.util.Accumulator#getWeight
       */
      private class AggregatorElementComparator
      implements Comparator<AggregatorElement>
      {
         /**
          * Compares two <code>AggregatorElement</code>s, based upon their
          * contained <code>Accumulator</code>'s weights.  Lower weights sort
          * first.
          *
          * @param   o1
          *          AggregatorElement 1.
          * @param   o2
          *          AggregatorElement 2.
          *
          * @return  Negative value if o1's weight is lower than o2's;
          *          positive value if o2's weight is lower than o1's;
          *          0 if they have equal weights.
          */
         @Override
         public int compare(AggregatorElement o1, AggregatorElement o2)
         {
            int weight1 = o1.getAccumulator().getWeight();
            int weight2 = o2.getAccumulator().getWeight();
            
            return (weight1 - weight2);
         }
      }
   }
   
   /**
    * Provide description for GenericFrame instance.
    * 
    * @see  java.lang.Object#toString()
    */
   @Override
   public String toString()
   {
      return frameName + "#" + frameId;
   }

   /**
    * Cleanup frame resources when a frame reaches its end of life.
    */
   static public class Callback
   implements Finalizable
   {
      /** Frame which will be cleaned up. */
      private GenericFrame frame;

      /**
       * Construct an instance for the given frame.
       *
       * @param   frame
       *          The frame for cleanup.
       */
      Callback(GenericFrame frame)
      {
         this.frame = frame;
      }

      /**
       * This method is called when frame reaches the end of execute()
       * procedure. Effectively this means that the frame is no longer
       * reachable by the application. As a part of the cleanup processing,
       * the frame is removed from the screen if it is not hidden yet with
       * another mechanisms.
       * 
       * @see  com.goldencode.p2j.util.Finalizable#finished()
       */
      @Override
      public void finished()
      {
         frame.destroy();
      }

      /**
       * This method is unused.
       *
       * @see  com.goldencode.p2j.util.Finalizable#iterate()
       */
      @Override
      public void iterate()
      {
      }

      /**
       * This method is unused.
       *
       * @see  com.goldencode.p2j.util.Finalizable#retry()
       */
      @Override
      public void retry()
      {
      }
   }
   
   /**
    * A class which stores two instances of the enclosing class:  an old frame
    * and a new frame.  The old frame represents the current scope in use by a
    * shared frame.  The new frame will be used to represent the shared frame
    * within an external procedure about to be entered.  The full state (i.e.,
    * all instance variables) of the old frame is copied into the new frame
    * upon construction of an instance of this class.  When this object's
    * {@link #finished()} method is invoked, select state is copied back (see
    * {@link #finished()} for details.
    */
   static public class SharedFrameCleaner
   implements Finalizable
   {
      /** Frame proxy used for current scope */
      private GenericFrame oldFrame;
      
      /** Frame proxy used for new scope */
      private GenericFrame newFrame;
      
      /**
       * Constructor.  Stores arguments and copies state from old frame to new
       * frame.  Temporarily replaces old frame in the frame registry within
       * <code>LogicalTerminal</code> with the new frame.
       * <p>
       * All state of the old frame is copied to the new frame, EXCEPT the
       * following instance members:
       * <ul>
       * <li>batch
       * <li>configClass
       * <li>configInstance
       * <li>n2w
       * <li>n2n
       * <li>n2h
       * <li>widgets
       * <li>headers
       * <li>frame
       * <li>frameDef
       * </ul>
       * 
       * @param   oldFrame
       *          Frame proxy used for current scope.
       * @param   newFrame
       *          Frame proxy used for new scope.
       */
      SharedFrameCleaner(GenericFrame oldFrame, GenericFrame newFrame)
      {
         this.oldFrame = oldFrame;
         this.newFrame = newFrame;
         
         // initialize new frame's state, based on old frame
         newFrame.frameId                 = oldFrame.frameId;
         newFrame.needsPageElementCleanup = oldFrame.needsPageElementCleanup;
         newFrame.scopeValid              = oldFrame.scopeValid;
         newFrame.frameBuf                = oldFrame.frameBuf;
         newFrame.currentOp               = oldFrame.currentOp;
         newFrame.editLock                = oldFrame.editLock;
         newFrame.statsColumn             = oldFrame.statsColumn;
         newFrame.redirOut                = oldFrame.redirOut;
         newFrame.aggregateHandler        = oldFrame.aggregateHandler;
         newFrame.pendingUpDown           = oldFrame.pendingUpDown;
         newFrame.dynDown                 = oldFrame.dynDown;
         newFrame.dynColumn               = oldFrame.dynColumn;
         newFrame.dynRow                  = oldFrame.dynRow;
         newFrame.dynTitle                = oldFrame.dynTitle;
         newFrame.hadView                 = oldFrame.hadView;
         newFrame.textGroup               = oldFrame.textGroup;
         newFrame.scopeNesting            = oldFrame.scopeNesting;
         newFrame.shared                  = oldFrame.shared;
         newFrame.screenBufferCache       = oldFrame.screenBufferCache;
         newFrame.dynamicDown             = oldFrame.dynamicDown;
         newFrame.fieldGroup              = oldFrame.fieldGroup;
         newFrame.insideSetup             = oldFrame.insideSetup;
         newFrame.deferredAggregators     = oldFrame.deferredAggregators;
         newFrame.canHide                 = oldFrame.canHide;
         newFrame.afterDown               = oldFrame.afterDown;
         newFrame.afterUnderline          = oldFrame.afterUnderline;
         newFrame.hadCondition            = oldFrame.hadCondition;
         newFrame.triggerActive           = oldFrame.triggerActive;
         newFrame.finalCheck              = oldFrame.finalCheck;
         newFrame.silentFormatErrors      = oldFrame.silentFormatErrors;
         newFrame.formatFilter            = oldFrame.formatFilter;
         newFrame.flushed                 = oldFrame.flushed;
         newFrame.sameViewDataRedirected  = oldFrame.sameViewDataRedirected;
         
         // new frame must temporarily replace the old in LogicalTerminal's
         // registry, to ensure frame/widget lookups operate properly in new
         // frame's scope
         LogicalTerminal.replaceFrame(newFrame);
      }
      
      /**
       * This method is called when frame reaches the end of execute()
       * procedure. The configuration/state which was used in the newer scope
       * will be copied back to the original frame (i.e., that used in the
       * outer scope which was open at the time this object was created).
       * Restore the old frame to <code>LogicalTerminal</code>'s registry.
       * <p>
       * All state of the new frame is copied back to the old frame, EXCEPT
       * the following instance members:
       * <ul>
       * <li>batch
       * <li>configClass
       * <li>configInstance
       * <li>n2w
       * <li>n2n
       * <li>n2h
       * <li>widgets
       * <li>headers
       * <li>frame
       * <li>frameDef
       * </ul>
       *  
       * @see  com.goldencode.p2j.util.Finalizable#finished()
       */
      @Override
      public void finished()
      {
         oldFrame.needsPageElementCleanup = newFrame.needsPageElementCleanup;
         oldFrame.scopeValid              = newFrame.scopeValid;
         oldFrame.frameBuf                = newFrame.frameBuf;
         oldFrame.currentOp               = newFrame.currentOp;
         oldFrame.editLock                = newFrame.editLock;
         oldFrame.statsColumn             = newFrame.statsColumn;
         oldFrame.redirOut                = newFrame.redirOut;
         oldFrame.aggregateHandler        = newFrame.aggregateHandler;
         oldFrame.pendingUpDown           = newFrame.pendingUpDown;
         oldFrame.dynDown                 = newFrame.dynDown;
         oldFrame.dynColumn               = newFrame.dynColumn;
         oldFrame.dynRow                  = newFrame.dynRow;
         oldFrame.dynTitle                = newFrame.dynTitle;
         oldFrame.hadView                 = newFrame.hadView;
         oldFrame.textGroup               = newFrame.textGroup;
         oldFrame.scopeNesting            = newFrame.scopeNesting;
         oldFrame.shared                  = newFrame.shared;
         oldFrame.screenBufferCache       = newFrame.screenBufferCache;
         oldFrame.dynamicDown             = newFrame.dynamicDown;
         oldFrame.fieldGroup              = newFrame.fieldGroup;
         oldFrame.insideSetup             = newFrame.insideSetup;
         oldFrame.deferredAggregators     = newFrame.deferredAggregators;
         oldFrame.canHide                 = newFrame.canHide;
         oldFrame.afterDown               = newFrame.afterDown;
         oldFrame.afterUnderline          = newFrame.afterUnderline;
         oldFrame.hadCondition            = newFrame.hadCondition;
         oldFrame.triggerActive           = newFrame.triggerActive;
         oldFrame.finalCheck              = newFrame.finalCheck;
         oldFrame.silentFormatErrors      = newFrame.silentFormatErrors;
         oldFrame.formatFilter            = newFrame.formatFilter;
         oldFrame.flushed                 = newFrame.flushed;
         oldFrame.sameViewDataRedirected  = newFrame.sameViewDataRedirected;
         
         // restore old frame in frame registry
         LogicalTerminal.replaceFrame(oldFrame);
         
         // Force pushing screen, since current shared frame instance
         // relies on its frame definition and needs it immediately.
         oldFrame.pushScreenDefinition(true);
      }

      /**
       * This method is unused.
       *
       * @see  com.goldencode.p2j.util.Finalizable#iterate()
       */
      @Override
      public void iterate()
      {
      }

      /**
       * This method is unused.
       *
       * @see  com.goldencode.p2j.util.Finalizable#retry()
       */
      @Override
      public void retry()
      {
      }
   }
}
